import pytest
import os
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
from aqua import Reader
from aqua.core.graphics import plot_single_map, plot_single_map_diff
from aqua.core.graphics import plot_vertical_profile, plot_vertical_profile_diff
from aqua.core.graphics import plot_timeseries, plot_seasonalcycle
from aqua.core.graphics import plot_maps, plot_maps_diff, plot_hovmoller
from aqua.core.graphics import plot_vertical_lines, plot_histogram
from aqua.core.graphics import plot_lat_lon_profiles, plot_seasonal_lat_lon_profiles
from conftest import DPI, LOGLEVEL

loglevel = LOGLEVEL

# Aliases with module scope for fixtures
@pytest.fixture(scope='module')
def reader_era5(era5_hpz3_monthly_reader):
    return era5_hpz3_monthly_reader

@pytest.fixture(scope='module')
def data_era5(era5_hpz3_monthly_data):
    return era5_hpz3_monthly_data

@pytest.fixture(scope='module')
def reader_era5_r100(era5_hpz3_monthly_r100_reader):
    return era5_hpz3_monthly_r100_reader

@pytest.fixture(scope='module')
def data_era5_r100(era5_hpz3_monthly_r100_data):
    return era5_hpz3_monthly_r100_data

@pytest.fixture(scope='module')
def fesom_r200_fixFalse_reader(fesom_test_pi_original_2d_r200_fixFalse_reader):
    return fesom_test_pi_original_2d_r200_fixFalse_reader

@pytest.fixture(scope='module')
def fesom_r200_fixFalse_data(fesom_test_pi_original_2d_r200_fixFalse_data):
    return fesom_test_pi_original_2d_r200_fixFalse_data

@pytest.fixture(scope='module')
def reader_ifs_tc():
    return Reader(model='IFS', exp='test-tco79', source='teleconnections', fix=True)

@pytest.fixture(scope='module')
def data_ifs_tc(reader_ifs_tc):
    return reader_ifs_tc.retrieve(var='skt')

@pytest.mark.graphics
class TestMaps:
    """Basic tests for the Single map functions"""

    def test_plot_single_map(self, tmp_path, fesom_r200_fixFalse_reader, fesom_r200_fixFalse_data):
        """
        Test the plot_single_map function
        """
        data_regrid = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data = data_regrid["sst"].isel(time=0)
        fig, ax = plot_single_map(data=plot_data,
                                  proj=ccrs.PlateCarree(),
                                  contour=False,
                                  extent=[-180, 180, -90, 90],
                                  nlevels=5,
                                  vmin=-2.0,
                                  vmax=30.0,
                                  sym=True,
                                  cmap='viridis',
                                  display=False,
                                  return_fig=True,
                                  transform_first=False,
                                  title='Test plot',
                                  cbar_label='Sea surface temperature [°C]',
                                  dpi=DPI,
                                  nxticks=5,
                                  nyticks=5,
                                  ticks_rounding=1,
                                  cbar_ticks_rounding=1,
                                  loglevel=loglevel)
        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_single_map.png')
        assert os.path.exists(tmp_path / 'test_plot_single_map.png')

    def test_plot_single_map_diff(self, tmp_path, fesom_r200_fixFalse_reader, fesom_r200_fixFalse_data):
        """
        Test the plot_single_map_diff function
        """
        data_regrid = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data = data_regrid["sst"].isel(time=0)
        data_regrid2 = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data2 = data_regrid2["sst"].isel(time=1)

        fig, ax = plot_single_map_diff(data=plot_data,
                                       data_ref=plot_data2,
                                       nlevels=5,
                                       vmin_fill=-5.0,
                                       vmax_fill=5.0,
                                       sym=False,
                                       vmin_contour=-2.0,
                                       vmax_contour=30.0,
                                       sym_contour=True,
                                       cmap='viridis',
                                       display=False,
                                       return_fig=True,
                                       title='Test plot',
                                       cbar_label='Sea surface temperature [°C]',
                                       dpi=DPI,
                                       nxticks=5,
                                       nyticks=5,
                                       gridlines=True,
                                       loglevel=loglevel)
        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_single_map_diff.png')
        assert os.path.exists(tmp_path / 'test_plot_single_map_diff.png')

    def test_plot_single_map_no_diff(self, fesom_r200_fixFalse_reader, fesom_r200_fixFalse_data):
        """
        Test the plot_single_map_diff function
        """
        data_regrid = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data = data_regrid["sst"].isel(time=0)
        plot_data2 = plot_data.copy()

        fig, ax = plot_single_map_diff(data=plot_data, return_fig=True,
                                       data_ref=plot_data2, loglevel=loglevel)

        assert fig is not None
        assert ax is not None

    def test_maps(self, tmp_path, fesom_r200_fixFalse_reader, fesom_r200_fixFalse_data):
        """Test plot_maps function"""
        data_regrid = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data = data_regrid["sst"].isel(time=0)
        data_regrid2 = fesom_r200_fixFalse_reader.regrid(fesom_r200_fixFalse_data)
        plot_data2 = data_regrid2["sst"].isel(time=1)
        fig = plot_maps(maps=[plot_data, plot_data2],
                        nlevels=5,
                        vmin=-2, vmax=30,
                        sym=False,
                        cmap='viridis',
                        title='Test plot',
                        titles=['Test plot 1', 'Test plot 2'],
                        cbar_label='Sea surface temperature [°C]',
                        nxticks=5, nyticks=6,
                        return_fig=True, loglevel=loglevel)
        assert fig is not None

        fig.savefig(tmp_path / 'test_plot_maps.png', dpi=DPI)
        assert os.path.exists(tmp_path / 'test_plot_maps.png')

        fig2 = plot_maps_diff(maps=[plot_data, plot_data],
                              maps_ref=[plot_data2, plot_data2],
                              nlevels=5,
                              vmin_fill=-2, vmax_fill=2,
                              vmin_contour=-2, vmax_contour=30,
                              sym=False, sym_contour=True,
                              cmap='viridis',
                              title='Test plot',
                              titles=['Test plot 1', 'Test plot 2'],
                              cbar_label='Sea surface temperature [°C]',
                              nxticks=5, nyticks=6,
                              return_fig=True, loglevel=loglevel)

        assert fig2 is not None

        fig2.savefig(tmp_path / 'test_plot_maps_diff.png', dpi=DPI)
        assert os.path.exists(tmp_path / 'test_plot_maps_diff.png')

    def test_maps_error(self):
        """Test plot_maps function with error"""
        with pytest.raises(ValueError):
            plot_maps(maps="test")

        with pytest.raises(ValueError):
            plot_maps_diff(maps="test", maps_ref="test")


@pytest.mark.graphics
class TestVerticalProfiles:
    """Basic tests for the Vertical Profile functions"""

    @pytest.fixture(autouse=True)
    def setup(self, reader_era5_r100, data_era5_r100):
        self.reader = reader_era5_r100
        self.data = data_era5_r100
        self.data = self.reader.regrid(self.data)
        
    def test_plot_vertical_profile(self, tmp_path):
        """Test the plot_vertical_profile function"""
        fig, ax = plot_vertical_profile(data=self.data['q'].isel(time=0).mean('lon'),
                                        var='q',
                                        vmin=-0.002,
                                        vmax=0.002,
                                        nlevels=8,
                                        return_fig=True,
                                        loglevel=loglevel)

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_vertical_profile.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_plot_vertical_profile.png')

    def test_plot_vertical_profile_diff(self, tmp_path):
        """Test the plot_vertical_profile_diff function"""
        fig, ax = plot_vertical_profile_diff(data=self.data['q'].isel(time=0).mean('lon'),
                                            data_ref=self.data['q'].isel(time=1).mean('lon'),
                                            var='q',
                                            vmin=-0.002,
                                            vmax=0.002,
                                            vmin_contour=-0.002,
                                            vmax_contour=0.002,
                                            add_contour=True,
                                            nlevels=8,
                                            return_fig=True,
                                            loglevel=loglevel)

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_vertical_profile_diff.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_plot_vertical_profile_diff.png')
        
@pytest.mark.graphics
class TestTimeseries:
    """Basic tests for the Timeseries functions"""

    def setup_method(self):
        """Setup method to retrieve data for testing"""
        model = 'IFS'
        exp = 'test-tco79'
        source = 'teleconnections'
        var = 'skt'
        self.reader = Reader(model=model, exp=exp, source=source, fix=True)
        data = self.reader.retrieve(var=var)

        self.t1 = data[var].isel(lat=1, lon=1)
        self.t2 = data[var].isel(lat=10, lon=10)

    def test_plot_timeseries(self, tmp_path):
        """Test the plot_timeseries function"""
        t1_yearly = self.reader.timmean(self.t1, freq='YS', center_time=True)
        t2_yearly = self.reader.timmean(self.t2, freq='YS', center_time=True)
        std_mon = self.t1.groupby('time.month').std('time')
        std_annual = t1_yearly.std(dim='time')

        data_labels = ['t1', 't2']

        fig, ax = plot_timeseries(monthly_data=[self.t1, self.t2], annual_data=[t1_yearly, t2_yearly],
                                  ref_monthly_data=self.t1,
                                  ref_annual_data=t1_yearly,
                                  std_monthly_data=std_mon,
                                  std_annual_data=std_annual,
                                  ref_label=data_labels[0],
                                  title='Temperature at two locations',
                                  data_labels=data_labels)

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_timeseries.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_plot_timeseries.png')

    def test_plot_seasonalcycle(self, tmp_path):
        """Test the plot_seasonalcycle function"""
        t1_seasonal = self.t1.groupby('time.month').mean('time')
        t2_seasonal = self.t2.groupby('time.month').mean('time')
        std_data = self.t1.groupby('time.month').std('time')

        fig, ax = plot_seasonalcycle(data=t1_seasonal, ref_data=t2_seasonal,
                                     std_data=std_data,
                                     title='Seasonal cycle of temperature at two locations',
                                     data_labels='t1',
                                     ref_label='t2',
                                     loglevel=loglevel)

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_seasonalcycle.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_seasonalcycle.png')

    def test_plot_ensemble(self, tmp_path):
        """Test the plot_timeseries function with ensemble data"""
        t1_yearly = self.reader.timmean(self.t1, freq='YS', center_time=True)
        t2_yearly = self.reader.timmean(self.t2, freq='YS', center_time=True)

        # Create ensemble mean and standard deviation (fake data for testing)
        ens_mon_mean = (t1_yearly + t2_yearly) / 2
        
        # simply using the mean here, the function will plot: mean +/- 2xSTD
        # NOTE: the STD is pointwise along time axis
        #ens_mon_std = ens_mon_mean.groupby('time.month').std(dim='time') 
        ens_mon_std = ens_mon_mean 
        ens_annual_mean = self.reader.timmean(ens_mon_mean, freq='YS', center_time=True)

        # NOTE: Similarly, we will use annual mean as STD for testing purposes
        # as done in the previous lines
        #ens_annual_std = ens_annual_mean.std(dim='time')
        ens_annual_std = ens_annual_mean.std(dim='time')

        fig, ax = plot_timeseries(monthly_data=[self.t1, self.t2],
                                  annual_data=[t1_yearly, t2_yearly],
                                  ens_monthly_data=ens_mon_mean,
                                  std_ens_monthly_data=ens_mon_std,
                                  ens_annual_data=ens_annual_mean,
                                  std_ens_annual_data=ens_annual_std,
                                  ens_label='Ensemble mean',
                                  title='Ensemble mean temperature at two locations')

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_ensemble.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_plot_ensemble.png')


@pytest.mark.graphics
class TestHovmoller:
    """Basic tests for the Hovmoller functions"""

    def setup_method(self):
        model = 'IFS'
        exp = 'test-tco79'
        source = 'teleconnections'
        var = 'skt'
        self.reader = Reader(model=model, exp=exp, source=source, fix=False)
        data = self.reader.retrieve(var=var)

        self.data = data[var]

    def test_plot_hovmoller(self, tmp_path):
        """Test the plot_hovmoller function"""
        fig2, ax2 = plot_hovmoller(data=self.data,
                                   return_fig=True,
                                   cmap='RdBu_r',
                                   invert_axis=True,
                                   invert_time=True,
                                   cbar_label='test-label',
                                   nlevels=10,
                                   sym=True,
                                   loglevel=loglevel)

        assert fig2 is not None
        assert ax2 is not None

        fig2.savefig(tmp_path / 'test_hovmoller2.png', dpi=DPI)
        assert os.path.exists(tmp_path / 'test_hovmoller2.png')

        fig, _ = plot_hovmoller(data=self.data,
                                 return_fig=True,
                                 contour=False,
                                 cmap='RdBu_r',
                                 invert_time=True,
                                 loglevel=loglevel)

        assert fig is not None

        fig.savefig(tmp_path / 'test_hovmoller3.png', dpi=DPI)

        assert os.path.exists(tmp_path / 'test_hovmoller3.png')

    def test_plot_hovmoller_error(self):

        with pytest.raises(TypeError):
            plot_hovmoller(data='test')


@pytest.mark.graphics
class TestVerticalLines:
    """Basic tests for the Vertical Line functions"""

    @pytest.fixture(autouse=True)
    def setup(self, reader_era5, data_era5):
        self.reader = reader_era5
        self.data = data_era5['q'].isel(time=0, cells=0)
        
    def test_plot_vertical_lines(self, tmp_path):
        """Test the plot_vertical_lines function"""
        fig, ax = plot_vertical_lines(data=self.data,
                                      ref_data=self.data * 0.8,
                                      lev_name='plev',
                                      labels=['test'],
                                      ref_label='ref',
                                      title='Test vertical line',
                                      return_fig=True,
                                      invert_yaxis=True,
                                      loglevel=loglevel)

        assert fig is not None
        assert ax is not None

        fig.savefig(tmp_path / 'test_plot_vertical_lines.png', dpi=DPI)

        # Check the file was created
        assert os.path.exists(tmp_path / 'test_plot_vertical_lines.png')

@pytest.mark.graphics
class TestLatLonProfiles:
    """Basic tests for the lat_lon_profiles function"""

    @pytest.fixture(autouse=True)
    def setup(self, data_ifs_tc):
        """Setup method to retrieve data for testing"""
        data = data_ifs_tc
        
        self.lat_profile = data['skt'].mean(dim='lon')
        self.lon_profile = data['skt'].mean(dim='lat')

    def _save_and_check(self, fig, tmp_path, filename):
        """Helper to save figure and verify file exists"""
        filepath = tmp_path / filename
        fig.savefig(filepath)
        assert os.path.exists(filepath)

    def test_plot_lat_lon_profiles_single(self, tmp_path):
        """Test plot_lat_lon_profiles with single DataArray"""
        fig, ax = plot_lat_lon_profiles(data=self.lat_profile.isel(time=0),
                                        title='Latitude profile test',
                                        data_labels=['Test data'],
                                        loglevel=loglevel)

        assert fig is not None
        assert ax is not None
        self._save_and_check(fig, tmp_path, 'test_lat_profile.png')

    def test_plot_lat_lon_profiles_multiple(self, tmp_path):
        """Test plot_lat_lon_profiles with multiple DataArrays"""
        data_list = [self.lat_profile.isel(time=0), 
                     self.lat_profile.isel(time=1)]
        
        fig, ax = plot_lat_lon_profiles(data=data_list,
                                        data_labels=['Time 0', 'Time 1'],
                                        title='Multiple latitude profiles',
                                        loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        self._save_and_check(fig, tmp_path, 'test_lat_profiles_multiple.png')

    def test_plot_lat_lon_profiles_with_ref(self, tmp_path):
        """Test plot_lat_lon_profiles with reference data"""
        data = self.lat_profile.isel(time=0)
        ref = self.lat_profile.isel(time=1)
        ref_std = self.lat_profile.std(dim='time')
        
        fig, ax = plot_lat_lon_profiles(data=data,
                                        ref_data=ref,
                                        ref_std_data=ref_std,
                                        data_labels=['Data'],
                                        ref_label='Reference',
                                        title='Profile with reference',
                                        loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        self._save_and_check(fig, tmp_path, 'test_lat_profile_with_ref.png')

    def test_plot_lon_profile(self, tmp_path):
        """Test plot_lat_lon_profiles with longitude profile"""
        fig, ax = plot_lat_lon_profiles(data=self.lon_profile.isel(time=0),
                                        title='Longitude profile test',
                                        loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        self._save_and_check(fig, tmp_path, 'test_lon_profile.png')

    def test_plot_lat_lon_profiles_no_spatial_coords(self, tmp_path):
        """Test plot_lat_lon_profiles with DataArray without spatial coordinates"""
        data_no_coords = xr.DataArray(
            np.random.rand(10),
            dims=['time'],
            coords={'time': range(10)}
        )

        fig, ax = plot_lat_lon_profiles(data=data_no_coords,
                                        title='No spatial coordinates test',
                                        loglevel=loglevel)

        assert fig is not None
        assert ax is not None
        assert len(ax.lines) == 0  # No lines should be plotted
        self._save_and_check(fig, tmp_path, 'test_lat_profile_no_coords.png')

@pytest.mark.graphics
class TestSeasonalMeans:
    """Basic tests for the Seasonal Means functions"""

    @pytest.fixture(autouse=True)
    def setup(self, data_ifs_tc):
        """Setup method to retrieve data for testing"""
        data = data_ifs_tc
        
        self.data_seasonal = data['skt'].mean(dim='lon')
        
        # Create seasonal data for DJF, MAM, JJA, SON
        # Using simple time slicing for testing purposes
        self.djf = self.data_seasonal.isel(time=slice(0, 3)).mean(dim='time')
        self.mam = self.data_seasonal.isel(time=slice(3, 6)).mean(dim='time')
        self.jja = self.data_seasonal.isel(time=slice(6, 9)).mean(dim='time')
        self.son = self.data_seasonal.isel(time=slice(9, 12)).mean(dim='time')

    def test_plot_seasonal_lat_lon_profiles(self, tmp_path):
        """Test plot_seasonal_lat_lon_profiles function"""
        
        seasonal_data = [self.djf, self.mam, self.jja, self.son]
        
        fig, axs = plot_seasonal_lat_lon_profiles(seasonal_data=seasonal_data,
                                                  title='Seasonal Profiles Test',
                                                  data_labels=['Test data'],
                                                  loglevel=loglevel)
        
        assert fig is not None
        assert axs is not None
        assert len(axs) == 4
        

        fig.savefig(tmp_path / 'test_seasonal_profiles.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_seasonal_profiles.png')

    def test_plot_seasonal_lat_lon_profiles_with_ref(self, tmp_path):
        """Test plot_seasonal_lat_lon_profiles with reference data"""
        
        seasonal_data = [self.djf, self.mam, self.jja, self.son]
        # Use slightly modified data as reference
        ref_data = [self.djf * 0.95, self.mam * 0.95, 
                    self.jja * 0.95, self.son * 0.95]
        ref_std = [self.djf.std(), self.mam.std(), 
                   self.jja.std(), self.son.std()]
        
        fig, axs = plot_seasonal_lat_lon_profiles(seasonal_data=seasonal_data,
                                                  ref_data=ref_data,
                                                  ref_std_data=ref_std,
                                                  title='Seasonal Profiles with Reference',
                                                  data_labels=['Data'],
                                                  ref_label='Reference',
                                                  loglevel=loglevel)
        
        assert fig is not None
        assert axs is not None
        
        fig.savefig(tmp_path / 'test_seasonal_profiles_with_ref.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_seasonal_profiles_with_ref.png')

    def test_plot_seasonal_lat_lon_profiles_multiple(self, tmp_path):
        """Test plot_seasonal_lat_lon_profiles with multiple models"""
        
        # Create second set of data (slightly different)
        djf2 = self.djf * 1.05
        mam2 = self.mam * 1.05
        jja2 = self.jja * 1.05
        son2 = self.son * 1.05
        
        seasonal_data = [[self.djf, djf2], [self.mam, mam2], 
                        [self.jja, jja2], [self.son, son2]]
        
        fig, axs = plot_seasonal_lat_lon_profiles(seasonal_data=seasonal_data,
                                                  title='Multiple Models Seasonal Profiles',
                                                  data_labels=['Model 1', 'Model 2'],
                                                  loglevel=loglevel)
        
        assert fig is not None
        assert axs is not None
        
        fig.savefig(tmp_path / 'test_seasonal_profiles_multiple.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_seasonal_profiles_multiple.png')

    def test_plot_seasonal_lat_lon_profiles_error(self):
        """Test plot_seasonal_lat_lon_profiles with invalid input"""
        
        # Test with wrong number of seasons
        with pytest.raises(ValueError):
            plot_seasonal_lat_lon_profiles(seasonal_data=[self.djf, self.mam])
        
        # Test with non-list input
        with pytest.raises(ValueError):
            plot_seasonal_lat_lon_profiles(seasonal_data=self.djf)

    def test_plot_seasonal_lat_lon_profiles_with_none_ref_std(self, tmp_path):
        """Test plot_seasonal_lat_lon_profiles with None in ref_std_data"""
        
        seasonal_data = [self.djf, self.mam, self.jja, self.son]
        ref_data = [self.djf * 0.95, self.mam * 0.95, 
                    self.jja * 0.95, self.son * 0.95]
        
        # Introduce None in ref_std_data
        ref_std = [None, self.mam.std(), None, self.son.std()]
        
        fig, axs = plot_seasonal_lat_lon_profiles(seasonal_data=seasonal_data,
                                                ref_data=ref_data,
                                                ref_std_data=ref_std,
                                                title='Seasonal with None ref_std',
                                                data_labels=['Data'],
                                                ref_label='Reference',
                                                loglevel=loglevel)
        
        assert fig is not None
        assert axs is not None
        
        fig.savefig(tmp_path / 'test_seasonal_none_ref_std.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_seasonal_none_ref_std.png')

@pytest.mark.graphics
class TestHistogram:
    """Basic tests for the Histogram functions"""

    def setup_method(self):
        """Setup method to create histogram data for testing"""
        # Create simple histogram data with center_of_bin dimension
        bins = np.linspace(0, 10, 20)
        values = np.random.exponential(scale=2, size=20)
        
        self.hist_data = xr.DataArray(
            values,
            dims=['center_of_bin'],
            coords={'center_of_bin': bins}
        )
        self.hist_data.center_of_bin.attrs['units'] = 'm/s'
        self.hist_data.attrs['units'] = 'count'

    def test_plot_histogram_basic(self, tmp_path):
        """Test basic histogram plotting"""
        fig, ax = plot_histogram(data=self.hist_data,
                                 title='Test Histogram',
                                 loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        
        fig.savefig(tmp_path / 'test_histogram_basic.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_histogram_basic.png')

    def test_plot_histogram_multiple_with_ref(self, tmp_path):
        """Test histogram with multiple data and reference"""
        data_list = [self.hist_data, self.hist_data * 0.8]
        ref_data = self.hist_data * 1.2
        
        fig, ax = plot_histogram(data=data_list,
                                ref_data=ref_data,
                                data_labels=['Data 1', 'Data 2'],
                                ref_label='Reference',
                                smooth=True,
                                smooth_window=3,
                                xlogscale=True,
                                loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        assert len(ax.lines) == 3  # 2 data + 1 ref
        
        fig.savefig(tmp_path / 'test_histogram_multi_ref.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_histogram_multi_ref.png')

    def test_plot_histogram_no_center_of_bin(self, tmp_path):
        """Test histogram with data missing center_of_bin dimension"""
        bad_data = xr.DataArray(np.random.rand(10), dims=['time'])
        
        fig, ax = plot_histogram(data=bad_data, loglevel=loglevel)
        
        assert fig is not None
        assert ax is not None
        assert len(ax.lines) == 0  # No lines plotted
        
        fig.savefig(tmp_path / 'test_histogram_no_bins.png', dpi=DPI)
        plt.close(fig)
        assert os.path.exists(tmp_path / 'test_histogram_no_bins.png')



"""Module to transform coordinates of an Xarray object."""

import xarray as xr
from aqua.core.logger import log_configure, log_history
from .coord_utils import get_data_model, units_conversion_factor
from .coordidentifier import CoordIdentifier


class CoordTransformer:
    """
    Class to transform coordinates of an Xarray object.
    It aims at transforming the coordinates provided by the user into
    a standard format.
    """

    def __init__(self, data, loglevel="WARNING"):
        """
        Constructor of the CoordTransator class.

        Args:
            data (xr.Dataset or xr.DataArray): Xarray Dataset or DataArray object.
            loglevel (str, optional): Log level. Defaults to 'WARNING'.
        """
        if not isinstance(data, (xr.Dataset, xr.DataArray)):
            raise TypeError("data must be an Xarray Dataset or DataArray object.")
        self.loglevel = loglevel
        self.logger = log_configure(self.loglevel, "CoordTransformer")

        self.data = data

        self.src_coords = CoordIdentifier(
            data.coords, loglevel=loglevel
        ).identify_coords()
        self.tgt_coords = None
        self.gridtype = self._info_grid(data.coords)
        self.logger.info("Grid type: %s", self.gridtype)

    def _info_grid(self, coords):
        """
        Identify the grid type of the Xarray object.
        To be used to check if the axis direction has to be reversed.
        Args:
            coords (xr.Coordinates): Coordinates of the Xarray object.
        Returns:
            str: The grid type of the Xarray object.
            It can be Regular, Curvilinear or Unstructured.
        """

        lonname = self.src_coords.get("latitude")
        latname = self.src_coords.get("longitude")
        if lonname is None or latname is None:
            return "Unknown"

        lat = coords[latname.get("name")]
        lon = coords[lonname.get("name")]
        if lon.ndim == 2 and lat.ndim == 2:
            return "Curvilinear"
        if lon.dims != lat.dims:
            return "Regular"
        return "Unstructured"

    def transform_coords(self, name="aqua", flip_coords=True):
        """
        Transform the coordinates of the Xarray object.

        Args:
            tgt_coords (dict, optional): Target coordinates dictionary. Defaults to None.
            name (str, optional): Name of the target data model. Defaults to "aqua".
            flip_coords (bool, optional): Whether to flip coordinates if necessary. Defaults to True.
        Returns:
            xr.Dataset or xr.DataArray: The transformed dataset or dataarray.
        """

        if not isinstance(name, str):
            raise TypeError("name must be a string.")

        # multiple safety check
        self.logger.info("Target data model: %s", name)
        data_yaml = get_data_model(name)
        self.tgt_coords = data_yaml.get("data_model")
        outname = f"{data_yaml.get('name')} v{str(data_yaml.get('version'))}"

        if not isinstance(self.tgt_coords, dict):
            raise TypeError("tgt_coords must be a dictionary.")

        data = self.data
        for coord in self.tgt_coords:
            if coord in self.src_coords and self.src_coords[coord]:
                tgt_coord = self.tgt_coords[coord]
                src_coord = self.src_coords[coord]
                self.logger.info("Analysing coordinate: %s", coord)
                # self.logger.info("Transforming coordinate %s to %s", src_coord, tgt_coord)
                data = self.rename_coordinate(data, src_coord, tgt_coord)
                if flip_coords:
                    data = self.flip_coordinate(data, src_coord, tgt_coord)
                data = self.convert_units(data, src_coord, tgt_coord)
                data = self.assign_attributes(data, tgt_coord)
            else:
                self.logger.info(
                    "Coordinate %s not found in source coordinates.", coord
                )

        log_history(data, f"Coordinates fixed toward {outname} datamodel")

        return data

    def rename_coordinate(self, data, src_coord, tgt_coord):
        """
        Rename coordinate if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.
        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with renamed coordinate.
        """
        if src_coord["name"] != tgt_coord["name"]:
            original_coords = list(data.coords)
            self.logger.info(
                "Renaming coordinate %s to %s", src_coord["name"], tgt_coord["name"]
            )
            data = data.rename({src_coord["name"]: tgt_coord["name"]})
            log_history(
                data,
                f"Renamed coordinate {src_coord['name']} to {tgt_coord['name']} by datamodel",
            )

            # Ensure the AQUA dependent index is preserved
            if f"idx_{src_coord['name']}" in original_coords:
                index_name = f"idx_{src_coord['name']}"
                new_index_name = f"idx_{tgt_coord['name']}"
                self.logger.info("Renaming index %s to %s", index_name, new_index_name)
                data = data.rename({index_name: new_index_name})

            # unclear if this is fundamental
            # if tgt_coord['name'] in data.dims:
            #   self.logger.info("Preserving original dimension %s and index.", src_coord['name'])
            #   data = data.swap_dims({tgt_coord['name']: src_coord['name']})
            #   data = data.set_index({src_coord['name']: tgt_coord['name']})

            # data = data.rename({src_coord['name']: tgt_coord['name']})
            tgt_coord["bounds"] = f"{tgt_coord['name']}_bnds"
            data = self._rename_bounds(data, src_coord, tgt_coord)
        return data

    def _rename_bounds(self, data, src_coord, tgt_coord):
        """
        Rename bounds if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.

        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with renamed bounds.
        """
        if src_coord["bounds"] is not None:
            if src_coord["bounds"] in data:
                self.logger.info(
                    "Renaming bounds %s to %s", src_coord["bounds"], tgt_coord["bounds"]
                )
                data = data.rename({src_coord["bounds"]: tgt_coord["bounds"]})
                data[tgt_coord["name"]].attrs["bounds"] = tgt_coord["bounds"]
            else:
                self.logger.info("Bounds %s not found in data.", src_coord["bounds"])
        return data

    def flip_coordinate(self, data, src_coord, tgt_coord):
        """
        Flip coordinate if necessary.

        Args:
            data (xr.Dataset or xr.DataArray): The Xarray object.
            src_coord (dict): Source coordinate dictionary.
            tgt_coord (dict): Target coordinate dictionary.

        Returns:
            xr.Dataset or xr.DataArray: The Xarray object with possibly flipped coordinate.
        """
        if "stored_direction" not in tgt_coord:
            return data
        if tgt_coord["stored_direction"] not in ["increasing", "decreasing"]:
            raise ValueError(
                f"tgt direction must be 'increasing' or 'decreasing', not  {tgt_coord['stored_direction']}"
            )
        if src_coord["stored_direction"] not in ["increasing", "decreasing"]:
            self.logger.warning(
                "src direction is not 'increasing' or 'decreasing', but %s. Disabling reverse!",
                src_coord["stored_direction"],
            )
            return data
        if src_coord["stored_direction"] != tgt_coord["stored_direction"]:
            if self.gridtype == "Regular":
                self.logger.info(
                    "Flipping coordinate %s from %s to %s",
                    tgt_coord["name"],
                    src_coord["stored_direction"],
                    tgt_coord["stored_direction"],
                )
                data = data.isel({tgt_coord["name"]: slice(None, None, -1)})
                # add an attribute for regridder evalution
                data[tgt_coord["name"]].attrs["flipped"] = 1
                log_history(
                    data,
                    f"Flipped coordinate {tgt_coord['name']} from {src_coord['stored_direction']} to {tgt_coord['stored_direction']} by datamodel",
                )
            else:
                self.logger.info(
                    "Cannot flip coordinate %s. Grid type is %s.",
                    tgt_coord["name"],
                    self.gridtype,
                )
        return data

    def convert_units(self, data, src_coord, tgt_coord):
        """
        Convert units of the coordinate.
        """
        if "units" not in tgt_coord:
            self.logger.debug(
                "%s unit not found in target data model. Disabling unit conversion.",
                tgt_coord["name"],
            )
            return data
        if "units" not in src_coord:
            self.logger.warning(
                "%s unit not found source data model. Disabling unit conversion.",
                src_coord["name"],
            )
            return data
        if "units" not in data[tgt_coord["name"]].attrs:
            self.logger.warning(
                "%s unit not found in data. Disabling unit conversion.",
                tgt_coord["name"],
            )
            return data
        if src_coord["units"].lower() in ["degree north", "degree east"]:
            fixed_unit = src_coord["units"].replace("degree", "degrees")
            self.logger.debug(
                "Fixing units for coordinate %s from %s to %s",
                src_coord["name"],
                src_coord["units"],
                fixed_unit,
            )
            src_coord["units"] = fixed_unit
        if src_coord["units"] != tgt_coord["units"]:
            self.logger.info(
                "Converting units of coordinate %s from %s to %s",
                src_coord["name"],
                src_coord["units"],
                tgt_coord["units"],
            )
            factor = units_conversion_factor(src_coord["units"], tgt_coord["units"])
            if factor is None:
                self.logger.warning(
                    "Incompatible unit conversion for coordinate %s: %s -> %s. Skipping conversion.",
                    src_coord["name"],
                    src_coord["units"],
                    tgt_coord["units"],
                )
                return data
            if factor != 0:
                self.logger.info("Conversion factor is: %s ", factor)
                data = data.assign_coords(
                    {tgt_coord["name"]: data[tgt_coord["name"]] * factor}
                )
                tgt_coord["bounds"] = f"{tgt_coord['name']}_bnds"
                data = self._convert_bounds(data, src_coord, tgt_coord, factor)
                log_history(
                    data,
                    f"Converted units of coordinate {src_coord['name']} from {src_coord['units']} to {tgt_coord['units']} by datamodel",
                )
            data[tgt_coord["name"]].attrs["units"] = tgt_coord["units"]
        return data

    def _convert_bounds(self, data, src_coord, tgt_coord, factor):
        """
        Convert units bounds of the coordinate.
        """
        if "bounds" not in tgt_coord:
            return data
        if tgt_coord["bounds"] in data:
            self.logger.info(
                "Converting bounds of coordinate %s from %s to %s",
                src_coord["name"],
                src_coord["units"],
                tgt_coord["units"],
            )
            data[tgt_coord["bounds"]] = data[tgt_coord["bounds"]] * factor
            data[tgt_coord["bounds"]].attrs["units"] = tgt_coord["units"]
        return data

    def assign_attributes(self, data, tgt_coord):
        """
        Assign attributes to the coordinate.
        """
        for key, value in tgt_coord.items():
            if key not in ["name", "units", "positive", "stored_direction", "bounds"]:
                if key not in data.coords[tgt_coord["name"]].attrs:
                    self.logger.debug(
                        "Adding attribute %s to coordinate %s", key, tgt_coord["name"]
                    )
                    data.coords[tgt_coord["name"]].attrs[key] = value
        return data


def counter_reverse_coordinate(data):
    """
    Flip back latitude if necessary
    """

    for coord in data.coords:
        if "flipped" in data.coords[coord].attrs:
            data = data.isel({coord: slice(None, None, -1)})
            del data.coords[coord].attrs["flipped"]
    return data



"""
This module contains the specific grid type builders for the different grid types.
"""

import numpy as np
from .basegridbuilder import BaseGridBuilder


class RegularGridBuilder(BaseGridBuilder):
    """
    Class to build regular lon-lat grid files.
    """
    logger_name = "RegularGridBuilder"
    requires_bounds = False

    def get_metadata(self, data):
        """
        Get metadata for the lon-lat grid based on the data size.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
        Returns:
            dict: Metadata for the lon-lat grid, including nlon, nlat, cdogrid, and aquagrid.
        """
        nlon = data['lon'].size
        nlat = data['lat'].size
        cdogrid = f"r{nlon}x{nlat}"
        aquagrid = f"r{int(36000 / nlon)}"
        if nlat % 2 == 1:
            aquagrid += "s"
        return {
            'nlon': nlon,
            'nlat': nlat,
            'cdogrid': cdogrid,
            'aquagrid': aquagrid,
            'remap_method': "con",
            'kind': 'regular'
        }


class GaussianRegularGridBuilder(BaseGridBuilder):
    """
    Class to build Gaussian regular lon-lat grid files.
    """
    logger_name = "GaussianRegularGridBuilder"
    requires_bounds = False

    def get_metadata(self, data):
        """
        Get metadata for the Gaussian regular lon-lat grid based on the data size.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
        Returns:
            dict: Metadata for the Gaussian regular lon-lat grid, including nlon, nlat, cdogrid, and aquagrid.
        """

        nlon = data['lon'].size
        nlat = data['lat'].size
        if data['lat'][0] > data['lat'][-1]:
            cdogrid = f"F{int(nlat / 2)}"
            aquagrid = f"F{int(nlat / 2)}"
        else:
            raise ValueError("Latitude values are not from North to South, cannot build Gaussian regular grid")
        return {
            'nlon': nlon,
            'nlat': nlat,
            'cdogrid': cdogrid,
            'aquagrid': aquagrid,
            'spectral_truncation': nlat - 1,
            'remap_method': "con",
            'kind': 'gaussianregular'
        }


class HealpixGridBuilder(BaseGridBuilder):
    """
    Class to build HEALPix grid files.
    """
    logger_name = "HEALpixGridBuilder"
    requires_bounds = False

    def get_metadata(self, data):
        """
        Get metadata for the HEALPix grid based on the data size.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
        Returns:
            dict: Metadata for the HEALPix grid, including nside, zoom, cdogrid, and aquagrid.
        """
        nside = np.sqrt(data['mask'].size / 12)
        zoom = int(np.log2(nside))
        return {
            'nside': nside,
            'zoom': zoom,
            'cdogrid': f"hp{int(nside)}_nested",
            'aquagrid': f"hpz{int(zoom)}_nested",
            'remap_method': "con",
            'cdo_options': '--force',
            'kind': 'healpix'
        }


class UnstructuredGridBuilder(BaseGridBuilder):
    """
    Class to build Unstructured grid files.
    """
    logger_name = "UnstructuredGridBuilder"
    requires_bounds = True
    bounds_error_message = "Data has no bounds, cannot create Unstructured grid"

    def get_metadata(self, data):
        """
        Get metadata for the Unstructured grid based on the data size.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
        Returns:
            dict: Metadata for the Unstructured grid, including nlon, nlat, cdogrid, and aquagrid.
        """
        if not self.grid_name:
            raise ValueError("Grid name is not set for UnstructuredGrid, please provide it")

        return {
            'aquagrid': self.grid_name,
            'cdogrid': None,
            'size': data['mask'].size,
            'remap_method': "con",
            'kind': 'unstructured',
        }


class CurvilinearGridBuilder(BaseGridBuilder):
    """
    Class to build Curvilinear grid files.
    """
    logger_name = "CurvilinearGridBuilder"
    requires_bounds = False

    def get_metadata(self, data):
        """
        Get metadata for the Curvilinear grid based on the data size.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
        Returns:
            dict: Metadata for the Curvilinear grid, including nlon, nlat, cdogrid, and aquagrid.
        """
        if not self.grid_name:
            raise ValueError("Grid name is not set for CurvilinearGrid, please provide it")

        if self.has_bounds(data):
            remap_method = "con"
        else:
            self.logger.warning("Bounds not found, using bilinear remapping for Curvilinear grid")
            remap_method = "bil"

        return {
            'aquagrid': self.grid_name,
            'cdogrid': None,
            'size': data['mask'].size,
            'remap_method': remap_method,
            'kind': 'curvilinear'
        }



"""Class for handling trend and detrending of xarray objects."""

import xarray as xr
import numpy as np
import pandas as pd
from aqua.core.logger import log_configure, log_history



class Trender:
    """
    A class to handle trend and detrending of xarray objects using polynomial fitting.
    """

    def __init__(self, loglevel: str = 'WARNING'):
        """
        Initialize the Trender class with optional default settings.

        Args:
            loglevel (str): Logging level. Default is 'WARNING'.
        """
        self.loglevel = loglevel
        self.logger = log_configure(self.loglevel, "Trender")

    def trend(self, data: xr.DataArray | xr.Dataset, dim: str = 'time',
              degree: int = 1, skipna: bool = False) -> xr.DataArray | xr.Dataset:
        """
        Estimate the trend of an xarray object using polynomial fitting.

        Args:
            data (DataArray or Dataset): The input data.
            dim (str): Dimension to apply trend along. Defaults to 'time'.
            degree (int): Degree of the polynomial. Defaults to 1.
            skipna (bool): Whether to skip NaNs. Defaults to False.

        Returns:
            DataArray or Dataset: The trend component.
        """
        return self._apply_trend_or_detrend(data, self._trend, dim, degree, skipna)

    def detrend(self, data: xr.DataArray | xr.Dataset, dim: str = 'time',
                degree: int = 1, skipna: bool = False) -> xr.DataArray | xr.Dataset:
        """
        Remove the trend from an xarray object using polynomial fitting.

        Args:
            data (DataArray or Dataset): The input data.
            dim (str): Dimension to apply detrend along. Defaults to 'time'.
            degree (int): Degree of the polynomial. Defaults to 1.
            skipna (bool): Whether to skip NaNs. Defaults to False.

        Returns:
            DataArray or Dataset: The detrended data.
        """
        return self._apply_trend_or_detrend(data, self._detrend, dim, degree, skipna)
    
    def coeffs(self, data: xr.DataArray | xr.Dataset, dim: str = 'time',
               degree: int = 1, skipna: bool = False, normalize: bool = False) -> xr.DataArray | xr.Dataset:
        """"
        Compute the polynomial coefficients for the trend.
        """

        return self._apply_trend_or_detrend(data, self._coeffs, dim, degree, skipna, normalize=normalize)

    def _coeffs(self, data: xr.DataArray | xr.Dataset, dim: str, degree: int, skipna: bool, normalize: bool) -> xr.DataArray | xr.Dataset:
        """
        Compute the polynomial coefficients for the trend, adjusted to the input data.

        Args:
            data (DataArray or Dataset): Input data.
            dim (str): Dimension to apply fit along.
            degree (int): Polynomial degree.
            skipna (bool): Whether to skip NaNs.
            normalize (bool): Whether to normalize coefficients for the 'time' dimension.
                              This applies a scaling factor based on the inferred frequency of the time dimension.
                              It is applied only if `dim` is 'time' and it is scaled to all degrees of the polynomial.

        Returns:
            DataArray or Dataset: Coefficients of the polynomial fit adjusted to the input data.
        """
        coeffs = data.polyfit(dim=dim, deg=degree, skipna=skipna)

        # keep consistency with datasets
        # coeffs.rename_vars({"polyfit_coefficients": data.name})

        # time axis are scaled to nanoseconds, which is not very user-friendly.
        # we try to adjust the coefficients to the input data frequency
        if dim == 'time' and normalize:
            self.logger.debug('Normalizing coefficients for time dimension.')
            # get the inferred frequency of the time dimension and convert to pandas offset
            time_values = data[dim].to_index()
            inferred_freq = time_values.inferred_freq
            self.logger.debug("Inferred frequency for 'time' dimension: %s", inferred_freq)
            if inferred_freq is None:
                raise ValueError("Inferred frequency for 'time' dimension is None. "
                                 "Ensure that the time dimension has a pandas compatible frequency.")
            offset = pd.tseries.frequencies.to_offset(inferred_freq)
            self.logger.debug("Offset for normalization: %s", offset)

            # offset cannot be converted to timedelta, so we use the mean of the time values
            delta = ((time_values + offset) - time_values).mean()
            factor = delta.value # convert to nanoscend

            self.logger.debug("Time normalization factor: %s", factor)

            # create an array of factor for each degree and convert to DataArray
            factor_scales = np.array([factor**(degree - i) for i in range(degree + 1)])
            self.logger.debug("Factor scales for polynomial degrees: %s", factor_scales)
            factor_da = xr.DataArray(factor_scales, dims="degree", coords={"degree": coeffs.coords["degree"]})

            # adjust the coefficients by the factor
            coeffs = coeffs * factor_da

        return coeffs.polyfit_coefficients

    def _apply_trend_or_detrend(self, data, func, dim, degree, skipna, **kwargs):
        """
        Internal dispatcher for trend/detrend logic.
        """
        action = func.__name__.capitalize() # Get the action name (Trend or Detrend)
        
        self.logger.info(
            "Applying %s with polynomial of order %d along '%s' dimension.", action, degree, dim
        )

        if isinstance(data, xr.DataArray):
            final = func(data=data, dim=dim, degree=degree, skipna=skipna, **kwargs)

        elif isinstance(data, xr.Dataset):
            selected_vars = [da for da in data.data_vars if dim in data[da].coords]
            final = data[selected_vars].map(func, keep_attrs=True,
                                            dim=dim, degree=degree, skipna=skipna, **kwargs)
        else:
            raise ValueError("Input must be an xarray DataArray or Dataset.")

        return log_history(final, f"{action} with polynomial of order {degree} along '{dim}' dimension")

    def _trend(self, data: xr.DataArray, dim: str, degree: int, skipna: bool) -> xr.DataArray:
        """
        Compute the trend component using polynomial fit.
        Taken from https://ncar.github.io/esds/posts/2022/dask-debug-detrend/
        According to the post, current implementation is not the most efficient one.

        Args:
            data (DataArray): Input data.
            dim (str): Dimension to apply fit along.
            degree (int): Polynomial degree.
            skipna (bool): Whether to skip NaNs.

        Returns:
            DataArray: Trend component.

        """
        coeffs = data.polyfit(dim=dim, deg=degree, skipna=skipna)
        return xr.polyval(data[dim], coeffs.polyfit_coefficients)

    def _detrend(self, data: xr.DataArray, dim: str, degree: int, skipna: bool) -> xr.DataArray:
        """
        Subtract the trend from the data.

        Args:
            data (DataArray): Input data.
            dim (str): Dimension to detrend along.
            degree (int): Polynomial degree.
            skipna (bool): Whether to skip NaNs.

        Returns:
            DataArray: Detrended data.
        """
        return data - self._trend(data, dim=dim, degree=degree, skipna=skipna)



"""
Module including time utilities for AQUA
"""
import math
import numpy as np
import pandas as pd
import xarray as xr
from pandas.tseries.frequencies import to_offset
from aqua.core.util.sci_util import generate_quarter_months, TRIPLET_MONTHS
from aqua.core.util.string import get_quarter_anchor_month
from aqua.core.logger import log_configure

def frequency_string_to_pandas(freq):
    """
    Convert a string from the AQUA convention to
    the usual pandas frequency standard

    Args:
        freq (str): The frequency string to convert.

    Returns:
        str: The converted frequency string, pandas compliant.
    """
    logger = log_configure('WARNING', 'frequency_string_to_pandas')

    trans = {
        'hourly': 'h',
        'daily': 'D',
        'weekly': 'W',
        'seasonal': 'QS-DEC',
        'monthly': 'MS', #this implies start of the month
        'annual': 'YS', #this implies start of the year
        'yearly': 'YS', #this implies start of the year
        'hour': 'h',
        'day': 'D',
        'pentad': '5d',
        'week': 'W',
        'month': 'MS',
        'year': 'YS',
        'hours': 'h',
        'days': 'D',
        'pentads': '5D',
        'weeks': 'W',
        'months': 'MS',
        'years': 'YS',
    }

    new_freq = trans.get(freq, freq)

    if freq in ['M', 'ME', 'Y', 'YE']:
        logger.warning('You are using a pandas frequency pointing at the end of a period, this can behave unexpectedly if you have subdaily data')

    return new_freq


def xarray_to_pandas_freq(xdataset: xr.Dataset | xr.DataArray):
    """
    Given a Xarray Dataset, estimate the time frequency and convert
    it as a Pandas frequency string

    Args:
        xdataset (xr.Dataset | xr.DataArray): The input xarray object.

    Returns:
        str: The inferred time frequency as a string, pandas compliant.
    """
    # to check if this is necessary
    timedelta = pd.Timedelta(xdataset.time.diff('time').mean().values)

    hours = math.floor(timedelta.total_seconds() / 3600)
    days = math.floor(hours / 24)
    months = math.floor(days / 28)  # Minimum month has around 28 days
    years = math.floor(days / 365)  # Assuming an average year has around 365 days

    # print([hours, days, months, years])

    if years >= 1:
        return f"{years}Y"
    elif months >= 1:
        return f"{months}MS"
    elif days >= 1:
        return f"{days}D"
    else:
        return f"{hours}h"


def pandas_freq_to_string(freq: str) -> str:
    """
    Convert a pandas frequency string to a more human-readable format.
    It also supports already human-readable formats.

    Args:
        freq (str): The pandas frequency string.

    Returns:
        str: The human-readable format of the frequency.
    """
    trans = {
        # Hourly
        'H': 'hourly',
        'h': 'hourly',
        '1h': 'hourly',
        'hourly': 'hourly',
        # Daily
        'D': 'daily',
        'd': 'daily',
        '1D': 'daily',
        'daily': 'daily',
        # Weekly
        'W': 'weekly',
        'weekly': 'weekly',
        # Seasonal
        'QS-DEC': 'seasonal',
        'Q': 'seasonal',
        # Monthly
        '1MS': 'monthly',
        'MS': 'monthly',
        'M': 'monthly',
        'ME': 'monthly',
        'mon': 'monthly',
        'monthly': 'monthly',
        # Annual
        '1Y': 'annual',
        'YS': 'annual',
        'Y': 'annual',
        'YE': 'annual',
        'yearly': 'annual',
        'years': 'annual',
        'annual': 'annual'
    }

    return trans.get(freq, freq)


def _find_end_date(start_date, offset):
    """Given a date and an offset in the form of pandas frequency
    return the expected end date of that period"""

    start_date = pd.Timestamp(start_date)
    end_date = start_date + to_offset(offset)
    # this because to_offset does not go to next month/year
    #if 'Y' in offset or 'M' in offset:
    #    end_date = end_date + pd.DateOffset(days=1)
    return end_date


def _generate_expected_time_series(start_date, frequency, time_period):
    """
    Given a start date, a pandas frequency and the data_frequency generate
    an expected time series
    """

    end_date = _find_end_date(start_date, time_period)
    time_series = pd.date_range(start=start_date, end=end_date, freq=frequency, inclusive='left')

    return time_series


def chunk_dataset_times(xdataset, resample_frequency, loglevel):
    """
    Common setup for chunk completeness checking.
    
    Returns:
        tuple: (data_frequency, chunks)
    """
    logger = log_configure(loglevel, 'timmean_chunking')

    # get frequency of the dataset. Expected to be regular!
    data_frequency = xarray_to_pandas_freq(xdataset)
    logger.debug('Data frequency detected as: %s', data_frequency)

    # convert offset
    pandas_period = to_offset(resample_frequency)

    normalized_dates = xdataset.time.to_index().to_period(pandas_period).to_timestamp()
    chunks = pd.date_range(start=normalized_dates[0],
                           end=normalized_dates[-1],
                           freq=resample_frequency)
    
    logger.info('%s chunks from %s to %s at %s frequency to be analysed', 
                len(chunks), chunks[0], 
                chunks[-1], resample_frequency)
    
    # if no chunks, no averages
    if len(chunks) == 0:
        raise ValueError(f'No chunks! Cannot compute average on {resample_frequency} period, not enough data')

    return data_frequency, chunks


def check_chunk_completeness(xdataset, resample_frequency='1D', loglevel='WARNING'):
    """
    Support function for timmean().
    Verify that all the chunks available in a dataset are complete given a
    fixed resample_frequency.
    Args:
        xdataset: The original dataset before averaging
        resample_frequency: the frequency on which we are planning to resample, based on pandas frequency
    Raise:
        ValueError if the there no available chunks
    Returns:
        A Xarray DataArray binary, 1 for complete chunks and 0 for incomplete ones, to be used by timmean()
    """

    data_frequency, chunks = chunk_dataset_times(xdataset, resample_frequency, loglevel)

    logger = log_configure(loglevel, 'timmean_chunk_completeness')

    check_completeness = []

    for chunk in chunks:
        end_date = _find_end_date(chunk, resample_frequency)
        logger.debug('Processing chunk from %s to %s', chunk, end_date)
        expected_timeseries = _generate_expected_time_series(chunk, data_frequency,
                                                             resample_frequency)
        expected_len = len(expected_timeseries)

        effective_len = len(xdataset.time[(xdataset['time'] >= chunk) &
                                          (xdataset['time'] < end_date)])
        logger.debug('Expected chunk length: %s, Effective chunk length: %s', expected_len, effective_len)
        if expected_len == effective_len:
            check_completeness.append(True)
        else:
            logger.warning('Chunk %s->%s for has %s elements instead of expected %s, timmean() will exclude this',
                           expected_timeseries[0], expected_timeseries[-1], effective_len, expected_len)
            check_completeness.append(False)

    # build the binary mask
    taxis = xdataset.time.resample(time=resample_frequency).mean()
    if sum(check_completeness) == 0:
        logger.warning('Not enough data to compute any average on %s period, returning empty array', resample_frequency)

    # Create a dict mapping chunk dates to completeness flag
    completeness_dict = {pd.Timestamp(chunk): is_complete for chunk, is_complete in zip(chunks, check_completeness)}

    # Align with the actual resampled time axis
    aligned_completeness = [completeness_dict.get(pd.Timestamp(t), False) for t in taxis.time.values]

    boolean_mask = xr.DataArray(aligned_completeness, dims=('time',), coords={'time': taxis.time})

    return boolean_mask


def check_seasonal_chunk_completeness(xdataset, resample_frequency='QS-DEC', loglevel='WARNING'):
    """
    Support function for timmean() to check seasonal chunk completeness.
    Verify that all seasonal (quarterly) chunks have complete months.
    
    For seasonal data (QS-DEC), this checks if each quarter has all 3 of its
    constituent months. Uses the same season definitions as select_season().
    
    Args:
        xdataset: The original dataset before averaging
        resample_frequency: the frequency on which we are planning to resample, 
                          expected to be 'QS-DEC' (or similar quarterly frequency)
        loglevel: logging level
    
    Raise:
        ValueError if there are no available chunks
    
    Returns:
        A Xarray DataArray binary, 1 for complete quarters and 0 for incomplete ones
    """
    data_frequency, chunks = chunk_dataset_times(xdataset, resample_frequency, loglevel)

    anchor_month = get_quarter_anchor_month(resample_frequency)

    logger = log_configure(loglevel, 'timmean_seasonal_completeness')

    # Generate quarter months: e.g. {'DEC': {'Q1': [12,1,2], 'Q2': [3,4,5], ...}}
    quarters_full = generate_quarter_months(anchor_month)
    quarter_months = quarters_full[anchor_month] # e.g. {'Q1': [12, 1, 2], 'Q2': [3, 4, 5], ...}

    # Build season_months: map from start_month 
    # e.g., {12: {12, 1, 2}, 3: {3, 4, 5}, 6: {6, 7, 8}, 9: {9, 10, 11}}
    season_months = {}
    for quarter_key in ['Q1', 'Q2', 'Q3', 'Q4']:
        months_list = quarter_months[quarter_key] # e.g. [12, 1, 2]
        start_month = months_list[0]
        season_months[start_month] = set(months_list)

    if 'D' in data_frequency or 'h' in data_frequency:
        logger.info('Data is sub-monthly (%s), first checking monthly completeness...', data_frequency)
        monthly_mask = check_chunk_completeness(xdataset, 
                                                resample_frequency='MS',
                                                loglevel=loglevel)
        # Get only complete months, use .resample().mean() to get time axis
        complete_month_times = xdataset.time.resample(time='MS').mean().time
        complete_month_times = complete_month_times.where(monthly_mask, drop=True)

        # Store complete month-year combinations as timestamps
        complete_month_timestamps = set(complete_month_times.to_index())
        logger.debug('Complete months available: %s', sorted(complete_month_timestamps))
    else:
        complete_month_timestamps = set(xdataset.time.to_index())
        logger.debug('Retrieved data frequency is monthly or coarser, using all months')

    check_completeness = []
    for chunk in chunks:
        end_date = _find_end_date(chunk, resample_frequency)
        logger.debug('Processing seasonal chunk from %s to %s', chunk, end_date)

        # Determine which season this is based on the start month
        start_month = chunk.month
        expected_months = season_months.get(start_month, set())

        # Get actual months present in this quarter period
        quarter_data = xdataset.time[(xdataset['time'] >= chunk) & 
                                     (xdataset['time'] < end_date)]

        if len(quarter_data) == 0:
            actual_months = set()
        else:
            # Extract which months we actually have
            actual_months = set(quarter_data.to_index().month)

        # Check if all expected months are present and complete
        has_all_months = expected_months.issubset(actual_months)

        # Additionally check that these months are complete
        if 'D' in data_frequency or 'h' in data_frequency:
            # Check completeness for the specific months within this quarter period
            # Generate the expected month timestamps for this quarter
            quarter_start = pd.Timestamp(chunk)
            quarter_end = pd.Timestamp(end_date)
            # Get all month starts within this quarter e.g. [2024-12-01, 2025-01-01, 2025-02-01]
            quarter_month_starts = pd.date_range(start=quarter_start, end=quarter_end, freq='MS', inclusive='left')
            # Check if all these specific month-year combinations are complete
            months_are_complete = quarter_month_starts.isin(complete_month_timestamps).all()
            is_complete = has_all_months and months_are_complete
        else:
            # For monthly data, just check presence
            is_complete = has_all_months

        if is_complete:
            check_completeness.append(True)
        else:
            season_name = mon_to_quarter_season_name(start_month)
            logger.warning(f"Seasonal chunk {chunk.strftime('%Y-%m')} ({season_name}) incomplete: expected months "
                           f"{sorted(expected_months)}, found {sorted(actual_months)}, timmean() will exclude this")
            check_completeness.append(False)

    if sum(check_completeness) == 0:
        logger.warning(f'Not enough data to compute any complete seasonal average on {resample_frequency} period, returning empty array')

    taxis = xdataset.time.resample(time=resample_frequency).mean()

    # Create a dict mapping chunk dates to completeness flag
    completeness_dict = {pd.Timestamp(chunk): is_complete for chunk, is_complete in zip(chunks, check_completeness)}

    # Align with the actual resampled time axis
    aligned_completeness = [completeness_dict.get(pd.Timestamp(t), False) for t in taxis.time.values]

    boolean_mask = xr.DataArray(aligned_completeness, dims=('time',), coords={'time': taxis.time})

    return boolean_mask


def time_to_string(time=None, format='%Y-%m-%d'):
    """Convert a time object to a string in the format YYYY-MM-DD

    Args:
        time: a time object, either a string, a datetime64 object or a pandas timestamp
        format: the format of the output string (YYYY-MM-DD by default, YYYYMMDD tested)

    Returns:
        A string in the format defined by the format argument

    Raises:
        ValueError if time is None or if time is not a supported type
    """
    if time is None:
        raise ValueError('time_to_string() requires a time argument')

    # Timestap for safer string handling
    if isinstance(time, str):
        time = pd.Timestamp(time)
    # Convert supported types into pandas.Timestamp
    if isinstance(time, (pd.Timestamp, np.datetime64)):
        ts = pd.to_datetime(time)
    else:
        raise ValueError('time_to_string() requires a time argument of type str, pd.Timestamp or np.datetime64')
    return ts.strftime(format)


def int_month_name(month, abbreviated=False):
    """
    Return month name from integer (1-12) if xarray functions cannot be used
    
    Args:
        month (int): The month as an integer (1-12).
        abbreviated (bool): Whether to return the abbreviated month name (default is False).

    Returns:
        str: The name of the month.
    """
    name = ["January", "February", "March", "April", "May", "June",
            "July", "August", "September", "October", "November", "December"][month - 1]
    return name[:3] if abbreviated else name


def mon_to_quarter_season_name(month):
    """
    Convert a month number (starting month of a quarter) to season abbreviation name.
    For QS-DEC, the quarter start months are 12, 3, 6, 9. Map them to their season based on TRIPLET_MONTHS
    
    Args:
        month (int): Month number (1-12), expected to be a quarter start month
    
    Returns:
        str: Season abbreviation (e.g., 'DJF', 'MAM', 'JJA', 'SON')
    """
    for season_name, months in TRIPLET_MONTHS.items():
        if months[0] == month:
            return season_name



"""Graphics utilities for Aqua."""
import math

import xarray as xr
import cartopy.util as cutil
import cartopy.crs as ccrs
import cartopy.mpl.ticker as cticker
import numpy as np
import healpy as hp
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
import matplotlib.path as mpath
import matplotlib.patches as mpatches

from aqua.core.logger import log_configure
from .sci_util import check_coordinates
from .string import unit_to_latex


def add_cyclic_lon(da: xr.DataArray):
    """
    Add a cyclic longitude point to a DataArray using cartopy
    and preserving the DataArray as data structure.

    It assumes that the longitude coordinate is named 'lon' and
    the latitude coordinate is named 'lat'.

    Args:
        da (xarray.DataArray): Input data array with longitude coordinate

    Returns:
        xarray.DataArray: The same input data array with the cyclic point added along longitude
    """
    if not isinstance(da, xr.DataArray) or da is None:
        raise ValueError("Input must be an xarray.DataArray object.")

    # Support both lon and longitude names
    lon_name, lat_name = coord_names(da)

    cyclic_da, cyclic_lon = cutil.add_cyclic_point(da, coord=da[lon_name])

    # update the longitude coordinate with cyclic longitude
    new_da = xr.DataArray(cyclic_da, dims=da.dims)
    new_da = new_da.assign_coords(lon=cyclic_lon)
    new_da = new_da.assign_coords(lat=da[lat_name])

    # Add old attributes to the new DataArray
    new_da.attrs = da.attrs

    return new_da


def plot_box(num_plots=0):
    """
    Evaluate the number of rows and columns for a plot
    based on the number of plots to be plotted.

    Args:
        num_plots (int): Number of plots to be plotted.

    Returns:
        num_rows (int): Number of rows for the plot.
        num_cols (int): Number of columns for the plot.

    Raises:
        ValueError: If the number of plots is 0.
    """
    if num_plots == 0:
        raise ValueError('Number of plots must be greater than 0.')

    num_cols = math.ceil(math.sqrt(num_plots))
    num_rows = math.ceil(num_plots / num_cols)

    return num_rows, num_cols


def minmax_maps(maps: list):
    """
    Find the minimum and maximum values of the maps values
    for a list of maps.

    After finding the minimum and maximum values,
    values are made simmetric around 0.

    Args:
        regs (list): List of maps.

    Returns:
        vmin (float): Minimum value of the colorbar.
        vmax (float): Maximum value of the colorbar.
    """

    minmax = (min([map.min().values for map in maps if map is not None]),
              max([map.max().values for map in maps if map is not None]))

    vmin = minmax[0]
    vmax = minmax[1]

    # Make values simmetric around 0
    absmax = max(abs(vmin), abs(vmax))
    vmin = -absmax
    vmax = absmax

    return vmin, vmax


def evaluate_colorbar_limits(maps: list, sym: bool = True):
    """
    Evaluate the minimum and maximum values of the colorbar

    Args:
        maps (list):     List of DataArrays.
        sym (bool, opt): If True, the colorbar is symmetrical around 0.

    Returns:
        vmin (float): Minimum value of the colorbar.
        vmax (float): Maximum value of the colorbar.
    """
    if maps is None:
        raise ValueError('DataArrays must be specified.')

    if sym:
        vmin, vmax = minmax_maps(maps)
    else:
        vmin = min([map.min().values for map in maps if map is not None])
        vmax = max([map.max().values for map in maps if map is not None])

    return vmin, vmax


def cbar_get_label(data: xr.DataArray, cbar_label: str = None,
                   loglevel='WARNING'):
    """
    Evaluate the colorbar label.

    Args:
        data (xarray.DataArray): Input data array.
        cbar_label (str, opt):   Colorbar label.
        loglevel (str, opt):     Log level.

    Returns:
        cbar_label (str): Colorbar label.
    """
    logger = log_configure(loglevel, 'cbar get label')

    if cbar_label is None:
        cbar_label = getattr(data, 'long_name', None)
        if cbar_label is None:
            cbar_label = getattr(data, 'short_name', None)
        if cbar_label is None:
            cbar_label = getattr(data, 'shortName', None)
        logger.debug("Using %s as colorbar label", cbar_label)

        units = getattr(data, 'units', None)

        if units:
            cbar_label = f"{cbar_label} [{unit_to_latex(units)}]"
            logger.debug("Adding units to colorbar label")

    if cbar_label is None:
        logger.warning("No colorbar label found, please specify one with the cbar_label argument.")

    return cbar_label


def set_map_title(data: xr.DataArray, title: str = None,
                  put_units=False, set_units=None,
                  use_attr_name=None, put_model_name=True,
                  put_exp_name=True, skip_varname=False,
                  loglevel='WARNING'):
    """
    Evaluate the map title.

    Args:
        data (xarray.DataArray):   Input data array.
        title (str, opt):          Explicit title override.
        put_units (bool):      Whether to include units in the title. Default True.
        set_units (str):       Manually override units in title.
        use_attr_name (str):   Specific attribute name to use as title.
        put_model_name (bool): Include 'AQUA_model' in title. Default True.
        put_exp_name (bool):   Include 'AQUA_exp' in title. Default True.
        skip_varname (bool):   Skip including variable name in title. Default False.
        loglevel (str, opt):       Logging level.

    Returns:
        title (str): Map title.
    """
    logger = log_configure(loglevel, 'set map title')

    if title:
        logger.debug("Explicit title provided: %s", title)
        return title
    
    title = ""
    varname = None

    if use_attr_name:
        varname = data.attrs.get(use_attr_name, None)
        if varname:
            logger.debug(f"Using title from specified attribute '{use_attr_name}': {varname}")
        else:
            logger.warning(f"Attribute '{use_attr_name}' not found in data attributes. Checking standard ones")

    if varname is None:
        # Getting the variables from the possible attrs in data
        for attr in ['long_name', 'short_name', 'shortName']:
            varname = data.attrs[attr] if attr in data.attrs else None
            if varname is not None:
                break

    units = set_units if set_units is not None else data.attrs.get('units', None)
    model = data.attrs["AQUA_model"] if 'AQUA_model' in data.attrs else None
    exp = data.attrs["AQUA_exp"] if 'AQUA_exp' in data.attrs else None
    time = data.time.values if 'time' in data.dims else None

    if varname and not skip_varname:
        title += varname
        if units and put_units:
            title += f" [{unit_to_latex(units)}]"
    if put_model_name and model:
            title += f" {model}"
    if put_exp_name and exp:
            title += f" {exp}"
    if time is not None:
        time = np.datetime_as_string(time, unit='h')
        title += f" {time}"
    if title == "":
        logger.warning("No title found, please specify one with the title argument.")
        title = None
    else:
        logger.debug(f"Using {title} as map title") 
    return title


def coord_names(data: xr.DataArray):
    """
    Get the names of the longitude and latitude coordinates.

    Args:
        data (xarray.DataArray): Input data array.

    Returns:
        tuple: (lon_name, lat_name) - Names of longitude and latitude coordinates, 
               or (None, None) if not found.
    """
    lon_name = None
    lat_name = None
    
    # Find longitude coordinate
    for lon_candidate in ['lon', 'longitude']:
        if lon_candidate in data.coords:
            lon_name = lon_candidate
            break
    
    # Find latitude coordinate  
    for lat_candidate in ['lat', 'latitude']:
        if lat_candidate in data.coords:
            lat_name = lat_candidate
            break
    
    return lon_name, lat_name


def ticks_round(ticks: list, round_to: int = None):
    """
    Round ticks to the nearest round_to value.

    Args:
        ticks (list):         Ticks value.
        round_to (int, opt):  Round to value.

    Returns:
        ticks (list):  Rounded tick value.
    """
    if round_to is None:
        # define round_to
        tick_span = ticks[1] - ticks[0]
        if tick_span <= 1:
            round_to = 2
        elif tick_span > 1 and tick_span <= 10:
            round_to = 1
        else:
            round_to = 0

    return np.round(ticks, round_to)


def set_ticks(data: xr.DataArray,
              fig: plt.figure,
              ax: plt.axes,
              nticks: tuple,
              lon_name: str,
              lat_name: str,
              ticks_rounding: int = None,
              proj=ccrs.PlateCarree(),
              loglevel='WARNING'):
    """
    Set the ticks of the map.

    Args:
        data (xr.DataArray): Data to plot.
        fig (matplotlib.figure.Figure): Figure.
        ax (matplotlib.axes._subplots.AxesSubplot): Axes.
        nticks (tuple): Number of ticks for x and y axes.
        lon_name (str): Name of the longitude coordinate.
        lat_name (str): Name of the latitude coordinate.
        ticks_rounding (int, optional): Number of digits to round the ticks.
        loglevel (str, optional): Log level. Defaults to 'WARNING'.

    Returns:
        matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot: Figure and axes.
    """
    logger = log_configure(loglevel, 'set_ticks')
    nxticks, nyticks = nticks

    try:
        lon_min = data[lon_name].values.min()
        lon_max = data[lon_name].values.max()
        (lon_min, lon_max), _ = check_coordinates(lon=(lon_min, lon_max),
                                                  lat=None,
                                                  default_coords={"lon_min": -180,
                                                                  "lon_max": 180,
                                                                  "lat_min": -90,
                                                                  "lat_max": 90},)
        logger.debug("Setting longitude ticks from %s to %s", lon_min, lon_max)
    except KeyError:
        logger.critical("No longitude coordinate found, setting default values")
        lon_min = -180
        lon_max = 180
    step = (lon_max - lon_min) / (nxticks - 1)
    xticks = np.arange(lon_min, lon_max + 1, step)
    xticks = ticks_round(ticks=xticks, round_to=ticks_rounding)
    logger.debug("Setting longitude ticks to %s", xticks)
    ax.set_xticks(xticks, crs=proj)
    lon_formatter = cticker.LongitudeFormatter()
    ax.xaxis.set_major_formatter(lon_formatter)

    # Latitude labels
    # Evaluate the latitude ticks
    try:
        lat_min = data[lat_name].values.min()
        lat_max = data[lat_name].values.max()
        _, (lat_min, lat_max) = check_coordinates(lat=(lat_min, lat_max),
                                                  lon=None,
                                                  default_coords={"lon_min": -180,
                                                                  "lon_max": 180,
                                                                  "lat_min": -90,
                                                                  "lat_max": 90},)
        logger.debug("Setting latitude ticks from %s to %s", lat_min, lat_max)
    except KeyError:
        logger.critical("No latitude coordinate found, setting default values")
        lat_min = -90
        lat_max = 90
    step = (lat_max - lat_min) / (nyticks - 1)
    yticks = np.arange(lat_min, lat_max + 1, step)
    yticks = ticks_round(ticks=yticks, round_to=ticks_rounding)
    logger.debug("Setting latitude ticks to %s", yticks)
    ax.set_yticks(yticks, crs=proj)
    lat_formatter = cticker.LatitudeFormatter()
    ax.yaxis.set_major_formatter(lat_formatter)

    return fig, ax

def generate_colorbar_ticks(vmin, vmax, nlevels=11, sym=False,
                            ticks_rounding=None, max_ticks=15, loglevel='WARNING'):
    """
    Generate and optionally round colorbar ticks for consistent and readable display.

    Args:
        vmin (float): Minimum colorbar value.
        vmax (float): Maximum colorbar value.
        nlevels (int): Number of desired levels. Default is 11.
        sym (bool): Whether the colorbar should be symmetric around zero.
        ticks_rounding (int, optional): Decimal places for rounding. Default is None.
        max_ticks (int): Max allowed ticks on the colorbar. Default is 15.
        loglevel (str): Logging level.

    Returns:
        np.ndarray: Array of colorbar tick positions.
    """
    logger = log_configure(loglevel, 'generate_colorbar_ticks')

    if sym:
        vmax = max(abs(vmin), abs(vmax))
        vmin = -vmax
        logger.debug(f"Using symmetric colorbar: {vmin=}, {vmax=}")

    cbar_ticks = np.linspace(vmin, vmax, nlevels + 1)

    # Reduce number of ticks if too many
    if len(cbar_ticks) > max_ticks:
        step = max(1, int(np.ceil(len(cbar_ticks) / max_ticks)))
        cbar_ticks = cbar_ticks[::step]
        # Ensure last tick is included
        if cbar_ticks[-1] != vmax:
            cbar_ticks = np.append(cbar_ticks, vmax)

    if ticks_rounding is not None:
        logger.debug(f"Rounding colorbar ticks to {ticks_rounding} decimals")
        cbar_ticks = ticks_round(cbar_ticks, ticks_rounding)

    return cbar_ticks

def apply_circular_window(ax, extent=None, apply_black_circle=False):
    """
    Apply a circular boundary mask to a Cartopy GeoAxes and set geographic extent
    to avoid the default rectangular plotting window with some projections.

    Args:
        ax (GeoAxes): Cartopy axes object to modify.
        extent (list or None): Geographic extent [west, east, south, north]. Default is Arctic region.

    Returns:
        ax (GeoAxes): Modified axes with circular boundary and extent.
    """
    if extent is None:
        extent = [-180, 180, 10, 90] # [west, east, south, north]

    # create circular boundary path
    theta = np.linspace(0, 2 * np.pi, 100)
    center, radius = [0.5, 0.5], 0.5
    verts = np.vstack([np.sin(theta), np.cos(theta)]).T
    circle = mpath.Path(verts * radius + center)

    # apply circle to axis
    ax.set_boundary(circle, transform=ax.transAxes)
    ax.set_extent(extent, crs=ccrs.PlateCarree())

    if apply_black_circle:
        # overlay a more visible black circle outline (drawn in axes coordinates)
        circle_patch = mpatches.Circle(
            center, radius=radius, transform=ax.transAxes,
            fill=False, color='darkgrey', linewidth=1.5, zorder=10)
        ax.add_patch(circle_patch)
    return ax

"""
Following functions are taken and adjusted from the easygems package,
on this repository:

https://github.com/mpimet/easygems/blob/main/easygems/healpix/__init__.py
"""
def get_nside(data):
    """
    Get the nside of a HEALPix map.

    Args:
        data (numpy.ndarray or xarray.DataArray): HEALPix map.
    
    Returns:
        int: nside of the HEALPix map.
    
    Raises:
        ValueError: If the input data is not a valid HEALPix map.
    """
    # Check if the input is a numpy array or xarray DataArray
    if not isinstance(data, (np.ndarray, xr.DataArray)): 
        raise ValueError("Input data must be a numpy array or xarray DataArray")

    if data.size == 0:  # Check for empty data
        raise ValueError("Invalid HEALPix map: data array is empty")
    
    npix = data.size
    if not hp.isnpixok(npix):
        raise ValueError(f"Invalid HEALPix map: npix={npix}")
    return hp.npix2nside(npix)
    
def get_npix(data):
    """
    Get the number of pixels in a HEALPix map based on the map data.

    Args:
        data (numpy.ndarray or xarray.DataArray): HEALPix map data.

    Returns:
        int: Number of pixels in the HEALPix map.
    """
    return hp.nside2npix(get_nside(data))
    
def healpix_resample(
        var,
        xlims=None,
        ylims=None,
        nx=None,
        ny=None,
        src_crs=None,
        method="nearest",
        nest=True,
        nside_out=None,
        loglevel="WARNING",
        ):
    """
    Resample a HEALPix map to a lat/lon grid.

    Args:
        var (xarray.DataArray): Input HEALPix map.
        xlims (tuple, optional): Longitude limits for the output grid.
        ylims (tuple, optional): Latitude limits for the output grid.
        nx (int, optional): Number of points in the x direction.
        ny (int, optional): Number of points in the y direction.
        src_crs (cartopy.crs.Projection, optional): Source coordinate reference system.
        method (str, optional): Resampling method ('nearest' or 'linear').
        nest (bool, optional): Whether to use nested HEALPix scheme.
        nside_out (int, optional): Output HEALPix nside.
        loglevel (str, optional): Log level.

    Returns:
        xarray.DataArray: Resampled data on a lat/lon grid.
    """

    logger = log_configure(loglevel, "healpix resample")

    nside = get_nside(var)
    if nside_out is None:
        nside_out = nside

    resolution_deg = 58.6 / nside_out
    if xlims is None:
        xlims = (-180, 180)
    if ylims is None:
        ylims = (-90, 90)

    if nx is None:
        nx = int((xlims[1] - xlims[0]) / resolution_deg)
    if ny is None:
        ny = int((ylims[1] - ylims[0]) / resolution_deg)

    if src_crs is None:
        src_crs = ccrs.Geodetic()

    # Compute grid centers
    dx = (xlims[1] - xlims[0]) / nx
    dy = (ylims[1] - ylims[0]) / ny
    xvals = np.linspace(xlims[0] + dx / 2, xlims[1] - dx / 2, nx)
    yvals = np.linspace(ylims[0] + dy / 2, ylims[1] - dy / 2, ny)
    xvals2, yvals2 = np.meshgrid(xvals, yvals)

    # Transform to lat/lon
    latlon = ccrs.PlateCarree().transform_points(
        src_crs, xvals2, yvals2, np.zeros_like(xvals2)
    )
    valid = np.all(np.isfinite(latlon), axis=-1)
    points = latlon[valid].T

    res = np.full(latlon.shape[:-1], np.nan, dtype=var.dtype)

    logger.debug("Resampling HEALPix map to lat/lon grid with %d x %d points", nx, ny)
    if method == "nearest":
        pix = hp.ang2pix(
            get_nside(var),
            theta=points[0],
            phi=points[1],
            nest=nest,
            lonlat=True,
        )
        if var.size < get_npix(var):
            if not isinstance(var, xr.DataArray):
                raise ValueError(
                    "Sparse HEALPix grids are only supported as xr.DataArray"
                )
            res[valid] = var.sel(cell=pix, method="nearest").where(
                lambda x: x.cell == pix
            )
        else:
            res[valid] = var[pix]

    elif method == "linear":
        lons, lats = hp.pix2ang(
            nside=get_nside(var),
            ipix=np.arange(len(var)),
            nest=nest,
            lonlat=True,
        )
        lons = (lons + 180) % 360 - 180

        valid_src = ((lons > points[0].min()) & (lons < points[0].max())) | (
            (lats > points[1].min()) & (lats < points[1].max())
        )

        res[valid] = griddata(
            points=np.asarray([lons[valid_src], lats[valid_src]]).T,
            values=var[valid_src],
            xi=(points[0], points[1]),
            method="linear",
            fill_value=np.nan,
            rescale=True,
        )

    result = xr.DataArray(res, coords=[("lat", yvals), ("lon", xvals)])
    result.attrs = getattr(var, "attrs", {}).copy()
    return result



#!/usr/bin/env python3
"""
AQUA analysis module for running diagnostics and handling configurations.
"""

import os
import sys
import subprocess
from importlib import resources as pypath

from aqua.core.util import create_folder, to_list
from aqua.core.configurer import ConfigPath


def run_command(cmd: str, log_file: str, logger=None) -> int:
    """
    Run a system command and capture the exit code, redirecting output to the specified log file.

    Args:
        cmd (str): Command to run.
        log_file (str): Path to the log file to capture the command's output.
        logger: Logger instance for logging messages.

    Returns:
        int: The exit code of the command.
    """
    try:
        log_file = os.path.expandvars(log_file)
        log_dir = os.path.dirname(log_file)
        create_folder(log_dir)

        with open(log_file, 'w', encoding='utf-8') as log:
            process = subprocess.run(
                cmd, shell=True, stdout=log, stderr=log, text=True, check=False
            )
            return process.returncode
    except (OSError, subprocess.SubprocessError) as e:
        if logger:
            logger.error(f"Error running command {cmd}: {e}")
        raise


def run_diagnostic(diagnostic: str, script_path: str, extra_args: str,
                   loglevel: str = 'INFO', logger=None, logfile: str = 'diagnostic.log'):
    """
    Run the diagnostic script with specified arguments.

    Args:
        diagnostic (str): Name of the diagnostic.
        script_path (str): Path to the diagnostic script.
        extra_args (str): Additional arguments for the script.
        loglevel (str): Log level to use.
        logger: Logger instance for logging messages.
        logfile (str): Path to the logfile for capturing the command output.
        cluster (str): Dask cluster address.
    """
    try:
        logfile = os.path.expandvars(logfile)
        create_folder(os.path.dirname(logfile))

        cmd = f"python {script_path} {extra_args} -l {loglevel} > {logfile} 2>&1"
        logger.info(f"Running diagnostic {diagnostic}")
        logger.debug(f"Command: {cmd}")

        process = subprocess.run(
            cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=False
        )

        if process.returncode != 0:
            logger.error(f"Error running diagnostic {diagnostic}: {process.stderr}")
        else:
            logger.info(f"Diagnostic {diagnostic} completed successfully.")
    except (OSError, subprocess.SubprocessError) as e:
        logger.error(f"Failed to run diagnostic {diagnostic}: {e}")


def _build_extra_args(**kwargs):
    """Build command line arguments from key-value pairs, skipping None values."""
    args = ""
    for flag, value in kwargs.items():
        if value is not None:
            args += f" --{flag} {value}"
    return args


def run_diagnostic_func(diagnostic: str, parallel: bool = False,
                        regrid: str = None, cli: dict = {},
                        diag_config=None, catalog=None, model='default_model', exp='default_exp',
                        source='default_source', source_oce=None,
                        startdate=None, enddate=None, realization=None,
                        output_dir='./output', loglevel='INFO',
                        logger=None, cluster=None):
    """
    Run the diagnostic and log the output, handling parallel processing if required.

    Args:
        diagnostic (str): Name of the diagnostic to run.
        parallel (bool): Whether to run in parallel mode.
        regrid (str): Regrid option.
        cli (dict): CLI definitions for the tools.
        diag_config (dict): Configuration dictionary loaded from YAML.
        catalog (str): Catalog name.
        model (str): Model name.
        exp (str): Experiment name.
        source (str): Source name.
        source_oce (str): Extra source name for ocean data when both are needed.
        startdate (str): Start date (YYYY-MM-DD). Defaults to None.
        enddate (str): End date (YYYY-MM-DD). Defaults to None.
        realization (str): Realization name. Defaults to None.
        output_dir (str): Directory to save output.
        loglevel (str): Log level for the diagnostic.
        logger: Logger instance for logging messages.
        cluster: Dask cluster scheduler address.
    """

    # Internal naming scheme:
    # diagnostic: the name of the wrapper metadiagnostic, e.g. atmosphere2d, climate_metrics, etc.
    # tool: the name of the individual command-line tool being run, e.g. biases, ecmean, etc.

    output_dir = os.path.expandvars(output_dir)
    create_folder(output_dir)

    # run individual tools in serial mode
    for tool, tool_config in diag_config.items():

        logger.info(f"Running tool: {tool} for diagnostic: {diagnostic}")
        
        cli_path = cli.get(tool)
        if cli_path is None:
            logger.error("CLI path for tool '%s' not found, skipping.", tool)
            continue
 
        if not os.path.exists(cli_path):
            logger.error("Script for tool '%s' not found at path: %s, skipping", tool, cli_path)
            continue

        outname = f"{output_dir}/{tool_config.get('outname', diagnostic)}"
        extra_args = tool_config.get('extra', "")

        # Build conditional arguments
        if regrid:
            extra_args += f" --regrid {regrid}"

        if parallel:
            nworkers = tool_config.get('nworkers')
            if nworkers is not None:
                extra_args += f" --nworkers {nworkers}"

        # This is needed for ECmean which uses multiprocessing
        if cluster and not tool_config.get('nocluster', False):
            extra_args += f" --cluster {cluster}"

        # Add standard arguments using helper function
        extra_args += _build_extra_args(
            catalog=catalog,
            realization=realization,
            startdate=startdate,
            enddate=enddate
        )

        if tool_config.get('source_oce', False) and source_oce:  # pass source_oce only if allowed by the diagnostic config file
            extra_args += f" --source_oce {source_oce}"

        cfgs = to_list(tool_config.get('config'))
        if not cfgs:
            logger.error(f"Config for tool '{tool}' not found, skipping.")
            continue

        for i, cfg in enumerate(cfgs, start=1):
            args = f"--model {model} --exp {exp} --source {source} --outputdir {outname} {extra_args} --config {cfg}"
            if len(cfgs) == 1:
                logfile = f"{output_dir}/{diagnostic}-{tool}.log"
            else:
                logfile = f"{output_dir}/{diagnostic}-{tool}-{i}.log"

            run_diagnostic(
                diagnostic=diagnostic,
                script_path=cli_path,
                extra_args=args,
                loglevel=loglevel,
                logger=logger,
                logfile=logfile
            )

def get_aqua_paths(*, args, logger):
    """
    Get both the AQUA path and the AQUA-analysis config path.

    Args:
        args: Command-line arguments.
        logger: Logger instance for logging messages.

    Returns:
        tuple: AQUA path and configuration path.
    """
    aqua_core_path = str(pypath.files('aqua.core'))
    aqua_diagnostics_path = str(pypath.files('aqua.diagnostics'))

    logger.debug(f"AQUA core path: {aqua_core_path}")
    logger.debug(f"AQUA diagnostics path: {aqua_diagnostics_path}")

    aqua_configdir = ConfigPath().configdir
    logger.debug(f"AQUA config dir: {aqua_configdir}")

    aqua_analysis_config_path = os.path.expandvars(args.config) if args.config and args.config.strip() else os.path.join(aqua_configdir, "analysis/config.aqua-analysis.yaml")
    if not os.path.exists(aqua_analysis_config_path):
        logger.error(f"Config file {aqua_analysis_config_path} not found.")
        sys.exit(1)
    logger.info(f"AQUA analysis config path: {aqua_analysis_config_path}")

    return aqua_core_path, aqua_diagnostics_path, aqua_configdir, aqua_analysis_config_path



"""
DataModel class for applying base coordinate transformations.
Provides a clean interface to CoordTransformer with caching.
"""
import xarray as xr
from aqua.core.logger import log_configure
from .coordtransformer import CoordTransformer
from .coord_utils import get_data_model


class DataModel:
    """
    Manage base data model transformations.
    
    Provides a clean interface to apply standard coordinate transformations
    based on AQUA data model specifications. Works independently from fixes files.
    
    This class handles:
    - Coordinate identification (via CoordIdentifier)
    - Standard coordinate transformations (rename, units, direction)
    - Dimension alignment
    - Attribute standardization
    
    Args:
        name (str): Data model name (e.g., "aqua"). Default is "aqua". Other can be added.
        loglevel (str): Log level for logging. Default is 'WARNING'.
        
    Example:
        >>> datamodel = DataModel(name="aqua", loglevel="DEBUG")
        >>> data = datamodel.apply(data)
    """

    def __init__(self, name: str = "aqua", loglevel: str = 'WARNING'):
        """
        Initialize DataModel.
        
        Args:
            name (str): Data model name
            loglevel (str): Log level
        """
        self.name = name
        self.loglevel = loglevel
        self.logger = log_configure(log_level=loglevel, log_name='DataModel')

        # Load data model config (cached)
        self.logger.debug("Initializing DataModel: %s", self.name)
        self.config = get_data_model(self.name)
    
    def apply(self, data: xr.Dataset, flip_coords=True) -> xr.Dataset:
        """
        Apply base data model transformations to dataset.
        
        Args:
            data (xr.Dataset): Input dataset
            flip_coords (bool): Whether to flip coordinate directions as per data model.
        
        Returns:
            xr.Dataset: Transformed dataset with standardized coordinates
        """
        self.logger.info("Applying data model: %s", self.name)
        return CoordTransformer(data, loglevel=self.loglevel).transform_coords(
            name=self.name, flip_coords=flip_coords
        )
    
    def get_config(self) -> dict:
        """
        Get the data model configuration.
        
        Returns:
            dict: Data model configuration dictionary
        """
        return self.config
    
    def __repr__(self):
        return f"DataModel(name='{self.name}', loglevel='{self.loglevel}')"



"""Test fixer functionality for Reader"""

import pytest
import numpy as np
from aqua import Reader
from aqua.core.fixer import EvaluateFormula

LOGLEVEL = 'DEBUG'

@pytest.fixture(scope='module')
def reader_ifs_tco79_long(ifs_tco79_long_reader):
    return ifs_tco79_long_reader

@pytest.fixture(scope='module')
def data_ifs_tco79_long(ifs_tco79_long_data):
    return ifs_tco79_long_data

@pytest.fixture(scope='module')
def reader_ifs_tco79_long_fixFalse(ifs_tco79_long_fixFalse_reader):
    return ifs_tco79_long_fixFalse_reader

@pytest.fixture(scope='module')
def data_ifs_tco79_long_fixFalse(ifs_tco79_long_fixFalse_data):
    return ifs_tco79_long_fixFalse_data

@pytest.mark.aqua
def test_fixer_ifs_long(data_ifs_tco79_long, data_ifs_tco79_long_fixFalse):
    """Test basic fixing"""

    ntime = [10, 20, 1000]  # points in time to be checked (includes 1 month jump)
    data0 = data_ifs_tco79_long_fixFalse  # Retrieve not fixed data
    ttr0 = data0.ttr[ntime, 0, 0]
    tas0 = data0['2t'][ntime, 5, 5]

    # Preliminary - did we read the correct values?
    assert pytest.approx(tas0.values) == [289.78904636, 284.16920838, 284.00620338]
    assert pytest.approx(ttr0.values) == [-6969528.64626286, -14032413.9597565, -9054387.41655567]

    # Now let's fix
    data1 = data_ifs_tco79_long  # Retrieve fixed data
    # This is the decumulated ttr
    tnlwrf = data1.tnlwrf[ntime, 0, 0]
    tas1 = data1['skt'][ntime, 5, 5]
    mtntrf = data1.mtntrf[ntime, 0, 0]
    mtntrf2 = data1.mtntrf2[ntime, 0, 0]

    # Did decumulation work ?
    assert pytest.approx(tnlwrf.values) == [-193.92693374, -194.7589371, -159.28750829]

    # Did we get a correct derived variable specified with paramId ?
    assert pytest.approx(tas1.values) == tas0.values + 1.0

    # Did unit conversion work?
    assert pytest.approx(mtntrf.values) == [-193.92693374, -194.7589371, -159.28750829]

    # Did we get a correct derived variable specified with grib shortname ?
    assert pytest.approx(mtntrf2.values) == mtntrf.values * 2

    # Metadata checks

    # History logged
    assert 'Variable var235, derived with 2t+1.0 by fixer' in tas1.attrs['history']

    # paramId and other attrs
    assert tas1.attrs['paramId'] == '235'

    assert mtntrf.attrs['paramId'] == '172179'
    assert mtntrf.attrs['units_fixed'] == 1
    assert mtntrf.attrs['units'] == 'W m**-2'

    assert mtntrf2.attrs['paramId'] == '999179'
    assert mtntrf2.attrs['units_fixed'] == 1
    assert mtntrf2.attrs['units'] == 'W m-2'  # these were coded by hand
    assert mtntrf2.attrs['long_name'] == 'Mean top net thermal radiation flux doubled'


@pytest.mark.aqua
def test_fixer_ifs_long_mindate():
    """Test fixing with a minimum date functionality"""

    reader = Reader(model="IFS", exp="test-tco79", source="long-mindate",
                    fix=True, loglevel=LOGLEVEL)
    data = reader.retrieve(var='2t')

    data1 = data['2t'].sel(time="2020-07-31T00:00")[1,1]
    data2 = data['2t'].sel(time="2020-08-01T00:00")[1,1]

    assert np.isnan(data1.values)
    assert not np.isnan(data2.values)
    assert 'mindate' in data['2t'].attrs
    assert data['2t'].attrs['mindate'] == '2020-08-01T00:00'


@pytest.mark.aqua
def test_fixer_ifs_names():
    """Check with fixer_name method"""

    reader = Reader(model="IFS", exp="test-tco79", source="short_masked", loglevel=LOGLEVEL)
    data = reader.retrieve(var=['2t'])
    assert data['2t'].attrs['donald'] == 'duck'

@pytest.mark.aqua
def test_fixer_ifs_disable():
    """Check with fixer_name: False method"""

    reader = Reader(model="IFS", exp="test-tco79", source="short_disable_fix", loglevel=LOGLEVEL)
    assert reader.fix is False

@pytest.mark.aqua
def test_fixer_ifs_timeshift():
    """Check fixer for timeshift with both timestep and pandas"""

    reader = Reader(model="IFS", exp="test-tco79", source="long-shift-timestep", loglevel=LOGLEVEL)
    data = reader.retrieve()
    assert data.time[0].values == np.datetime64('2020-01-19T00:00:00')

    reader = Reader(model="IFS", exp="test-tco79", source="long-shift-pandas", loglevel=LOGLEVEL)
    data = reader.retrieve()
    assert data.time[0].values == np.datetime64('2020-01-01T00:00:00')


@pytest.mark.aqua
def test_fixer_ifs_coords():
    """Check with fixer_name and coords block"""

    reader = Reader(model="IFS", exp="test-tco79", source="short_masked-coord-test", loglevel=LOGLEVEL)
    data = reader.retrieve()
    assert 'timepippo' in data.coords
    assert 'cellspippo' in data.dims


@pytest.mark.aqua
def test_fixer_fesom_coords():
    """Check with fixer_name and coords block"""

    reader = Reader(model="FESOM", exp="test-pi", source="original_3d_coord_fix", datamodel=False, loglevel=LOGLEVEL)
    data = reader.retrieve()
    assert 'level' in data.coords
    assert 'a lot of water' in data.level.attrs['units']


@pytest.mark.aqua
def test_fixer_fesom_names():
    """Check with fixer parent from fixer_name method"""

    reader = Reader(model="FESOM", exp="test-pi", source="original_2d_fix", loglevel=LOGLEVEL)
    data = reader.retrieve()
    assert data['mlotst125'].attrs['uncle'] == 'scrooge'

@pytest.mark.aqua
def test_fixer_deltat():
    """Check that output for deltat read from metadata and from fixes are the same"""
    
    reader1 = Reader(model='IFS', exp='test-tco79', source='long-deltat', loglevel=LOGLEVEL)
    data_metadata = reader1.retrieve(var='tnlwrf').isel(time=5)
    reader2 = Reader(model='IFS', exp='test-tco79', source='long', loglevel=LOGLEVEL)
    data_fixer = reader2.retrieve(var='tnlwrf').isel(time=5)
    assert data_metadata.equals(data_fixer)
    assert reader1.fixer.deltat == 3600
    assert reader2.fixer.deltat == 3600


@pytest.fixture
def data_2t_tp(era5_hpz3_monthly_data):
    return era5_hpz3_monthly_data

@pytest.mark.aqua
class TestEvaluateFormula:
    def test_evaluate_formula(self, data_2t_tp):
        formula = "2t -273.15"
        convert = EvaluateFormula(
            data=data_2t_tp, formula=formula,
            units='Celsius',
            short_name="2t_celsius",
            long_name='2t converted to Celsius').evaluate()

        assert convert.attrs['short_name'] == '2t_celsius'
        assert convert.attrs['long_name'] == '2t converted to Celsius'
        assert convert.attrs['units'] == 'Celsius'

        original_values = (data_2t_tp['2t'].isel(time=0) - 273.15).mean()
        expected_values = convert.isel(time=0).mean()
        assert np.allclose(original_values.values, expected_values.values)

    def test_complex_formula(self, data_2t_tp):
        formula = "((2t - 273.15)^2)/2 +  (tprate / 1000)"
        convert = EvaluateFormula(
            data=data_2t_tp, formula=formula,
            units='Celsius^2 * hPa',
            short_name="complex_calc",
            long_name='Complex calculation').evaluate()

        assert convert.attrs['short_name'] == 'complex_calc'
        assert convert.attrs['long_name'] == 'Complex calculation'
        assert convert.attrs['units'] == 'Celsius^2 * hPa'

        original_values = (((data_2t_tp['2t'].isel(time=0) - 273.15)**2)/2 + (data_2t_tp['tprate'].isel(time=0) / 1000)).mean()
        expected_values = convert.isel(time=0).mean()
        assert np.allclose(original_values.values, expected_values.values)

    def test_magnitude(self, data_2t_tp):
        """Test magnitude calculation"""
        formula = "(2t^2 + 2t^2)^0.5"
        convert = EvaluateFormula(
            data=data_2t_tp, formula=formula,
            units='K',
            short_name="magnitude_2t",
            long_name='Magnitude of 2t vector').evaluate()
        
        assert convert.attrs['short_name'] == 'magnitude_2t'
        assert convert.attrs['long_name'] == 'Magnitude of 2t vector'
        assert convert.attrs['units'] == 'K'
        original_values = ((data_2t_tp['2t'].isel(time=0)**2 + data_2t_tp['2t'].isel(time=0)**2)**0.5).mean()
        expected_values = convert.isel(time=0).mean()
        assert np.allclose(original_values.values, expected_values.values)

    def test_wrong_formula(self, data_2t_tp):
        """Test wrong parentheses handling"""
        formula = "(2t - 273.15)) + (tprate / 1000"
        with pytest.raises(ValueError):
            EvaluateFormula(data=data_2t_tp, formula=formula).evaluate()
        formula = "((2t - 273.15) + (tprate / 1000)"
        with pytest.raises(ValueError):
            EvaluateFormula(data=data_2t_tp, formula=formula).evaluate()
        with pytest.raises(KeyError):
            EvaluateFormula(data=data_2t_tp, formula="2t ++ tprate").evaluate()



"""Utilities for calendar and timestep calculations"""

import pandas as pd
from datetime import datetime
import os
import numpy as np

def date2str(dateobj):
    """
    Converts a date object to a date, time string representation

    Args:
        dateobj (datetime): datetime object

    Returns:
        date, time (str): Date and time as a strings
    """

    return dateobj.strftime('%Y%m%d'), dateobj.strftime('%H%M')

def date2yyyymm(dateobj):
    """
    Converts a date object to year, month string representation

    Args:
        dateobj (datetime): datetime object

    Returns:
        year, month (str): Date as year and month
    """

    return dateobj.strftime('%Y'), dateobj.strftime('%m')


def add_offset(data_start_date, startdate, offset, timestep):
    """
    Sets initial date based on an offset in steps (special for DE 6h case)

    Args:
        data_start_date (datetime.datetime): Data start date
        startdate (datetime.datetime): User-defined starting date
        offset (int): Offset to add to data start date
        timestep (int): Timestep to use (only H and D implemented)

    Returns:
        New starting date (str)
    """

    if int(offset) != 0:
        if timestep.upper() in ["H", "D"]:
            base_date = pd.Timestamp(str(data_start_date)) + pd.Timedelta(int(offset), unit=timestep)
        else:
            raise ValueError("Timestep not supported")

        startdate_obj = pd.Timestamp(str(startdate))
        if startdate_obj > base_date:
            base_date = startdate_obj

        return base_date.strftime('%Y%m%dT%H%M')
    else:
        return startdate


def check_dates(startdate, start_date, enddate, end_date):
    """
    Check if starting and ending dates are within given range

    Args:
        startdate (datetime.datetime): Starting date
        start_date (datetime.datetime): Data start date
        enddate (datetime.datetime): Ending date
        end_date (datetime.datetime): Data end date

    Raises:
        ValueError: If the given date is not within the data range
    """

    startdate_fmt = "%Y%m%dT%H%M" if 'T' in str(startdate) else "%Y%m%d"
    start_date_fmt = "%Y%m%dT%H%M" if 'T' in str(start_date) else "%Y%m%d"
    enddate_fmt = "%Y%m%dT%H%M" if 'T' in str(enddate) else "%Y%m%d"
    end_date_fmt = "%Y%m%dT%H%M" if 'T' in str(end_date) else "%Y%m%d"

    if datetime.strptime(str(startdate), startdate_fmt) < datetime.strptime(str(start_date), start_date_fmt):
        raise ValueError(f"Starting date {str(startdate)} is earlier than the data start at {str(start_date)}.")

    if datetime.strptime(str(enddate), enddate_fmt) > datetime.strptime(str(end_date), end_date_fmt):
        raise ValueError(f"End date {str(enddate)} is later than the data end at {str(end_date)}.")

    if datetime.strptime(str(startdate), startdate_fmt) > datetime.strptime(str(enddate), enddate_fmt):
        raise ValueError(f"Start date {str(startdate)} is later than the end date at {str(enddate)}.")

    if datetime.strptime(str(start_date), start_date_fmt) > datetime.strptime(str(end_date), end_date_fmt):
        raise ValueError(f"Data start date {str(start_date)} is later than the data end at {str(end_date)}.")


def shift_time_dataset(data):
    """
    Shift time of a dataset back one month
    Args:
        data (xarray.DataSet): The dataset to shift
    Returns:
        A revised xarray.DataSet
    """

    newtime = pd.DatetimeIndex(data.time.values) + pd.DateOffset(months=-1)
    return data.assign_coords(time=newtime)


def split_date(datestr, timedefault="0000"):
    """
    Splits a date in YYYYMMDD:HHMM format into date and time strings
    Args:
        datestr (str): The date string to split
        timedefault (str): Default for time if not provided
    Returns:
        date and time as str
    """

    dd = str(datestr).split('T')
    timestr = dd[1] if "T" in str(datestr) else timedefault
    return dd[0], timestr


def make_timeaxis(data_startdate, startdate, enddate, timestep=None,
                  savefreq=None, chunkfreq=None, shiftmonth=False, skiplast=False,
                  bridge_start_date=None, bridge_end_date=None):
    """
    Compute timeaxis and chunk start and end dates and indices.

    Args:
        data_startdate (datetime.datetime): Starting date of the dataset
        startdate (datetime.datetime): Starting date of the time axis
        enddate (datetime.datetime): Ending date of the time axis
        offset (int): An initial offset for steps (to be used e.g. for 6H data saved starting from step=6)
        timestep (str): Timestep. Can be one of h, D, M
        savefreq (str): Frequency at which the data are saved. Can be one of h, 6h, D, M
        chunkfreq (str): Frequency at which the data are to be chunked. Can be one of D, M, Y
        shiftmonth (bool): If True, fixes data accumulated at the end of the month. Default is False.
        skiplast (bool): If True, skips the last date. Default is False.
        bridge_start_date (datetime.datetime): Starting date of bridge data (optional)
        bridge_end_date (datetime.datetime): Ending date of bridge data (optional)

    Returns:
        A tuple containing:
            - timeaxis (pd.Series): The time axis
            - chunkstart_idx (int): The starting index of each chunk
            - chunkstart (pd.Timestamp): The start date of the first chunk
            - chunkend_idx (int): The last index of each chunk
            - chunkend (pd.Timestamp): The end date of each chunk
            - chunksize (int): The number of data points in each chunk
            - chunktype (int): The type of chunk, hpc: 0 or bridge: 1
    """

    # these are equivalent, unless specified different
    if savefreq is None:
        savefreq = timestep
    if timestep is None:
        timestep = savefreq
    if chunkfreq is None:
        chunkfreq = timestep

    if shiftmonth and savefreq != "M":
        raise ValueError("shiftmonth option requested but data are not saved at monthly frequency!")

    # compute offset
    offset = len(pd.date_range(str(data_startdate), str(startdate), freq=timestep)) - 1

    sdate = pd.Timestamp(str(startdate))
    edate = pd.Timestamp(str(enddate))

    if shiftmonth:  # We will need one month more
        edate = edate + pd.offsets.MonthBegin()

    if skiplast:
        edate = edate - pd.Timedelta(1, unit=timestep)

    dates = pd.date_range(sdate, edate, freq=timestep)
    idx = range(len(dates))
    ts = pd.Series(idx, index=dates)

    if timestep != savefreq:
        # the data are saved at a frequency different from the original timestep (eg. monthly)
        # do a preliminary resampling
        if shiftmonth:
            idx = ts.resample(savefreq).min().values
            ts = pd.Series(idx[1:], index=dates[idx[0:-1]])  # index of the next month but date of previous one
            idx = idx[0:-1]
        else:
            idx = ts.resample(savefreq).min().values
            ts = pd.Series(idx, index=dates[idx])

    tsr = ts.resample(chunkfreq)
    sidx = tsr.min().values
    eidx = tsr.max().values
    chunksize = tsr.count().values

    if shiftmonth:  # special case, data is accumulated at end of month
        sdate = dates[sidx] - pd.offsets.MonthBegin()
        edate = dates[eidx] - pd.offsets.MonthBegin()
    else:
        sdate = dates[sidx]
        edate = dates[eidx]

    if bridge_end_date:
        bridge_start = pd.Timestamp(str(bridge_start_date))
        bridge_end = pd.Timestamp(str(bridge_end_date))
        chunktype = np.where((sdate >= bridge_start) & (sdate <= bridge_end), 1, 0)
        chunktype_end = np.where((edate >= bridge_start) & (edate <= bridge_end), 1, 0)
    else:
        chunktype = np.zeros(len(dates), dtype=int)
        chunktype_end = np.zeros(len(dates), dtype=int)

    return {
        'timeaxis': dates[idx],
        'start_idx': sidx + offset,
        'start_date': sdate,
        'end_idx': eidx + offset,
        'end_date': edate,
        'size': chunksize,
        'type': chunktype,
        'type_end': chunktype_end  # this is used only for chunk alignment sanity check
    }


def todatetime(datestr):
    """
    Converts a date string to a datetime object

    Args:
        datestr (str): Date string

    Returns:
        datetime object
    """

    return pd.Timestamp(str(datestr))


def read_bridge_date(obj):
    """
    Reads the bridge end date from a file or string
    """
    
    if obj and obj != "complete" and os.path.isfile(obj):
        with open(obj, 'r') as file:
            date = file.read()
        date = pd.Timestamp(date.strip())
        return date.strftime('%Y%m%dT%H%M')
    else:
        return obj
    
def floor_datetime(dt, freq, output_format="%Y%m%dT%H%M"):
    """
    Floors a datetime object to the specified pandas frequency, 
    using `to_period` and special last string drop for better handling of non-fixed frequencies.
    
    Parameters:
        dt (datetime or pandas.Timestamp): The datetime object to be floored.
        freq (str): The frequency string (e.g., 'H', 'D', '6H', '1MS', '3YS').
         output_format (str): The format for output datetime as a string. Default is "%Y%m%dT%H%M".
        
    Returns:
        String with "%Y%m%dT%H%M" format (e.g. 19900101T000)
    """

    # safety checks
    if dt is None:
        return dt
    if dt == 'complete':
        return dt

    if not isinstance(dt, pd.Timestamp):
        dt = pd.Timestamp(str(dt))

    if freq in ['M', 'MS']:  # Floor to the first of the month
        dt = pd.Timestamp(dt.year, dt.month, 1)
    elif freq in ['Y', 'YS']:  # Floor to the first of the year
        dt = pd.Timestamp(dt.year, 1, 1)
    elif freq in ['ME', 'YE']:
        raise KeyError(f'Freq {freq} is not supported, please use {freq[0]}S')
    elif 'W' in freq: 
        raise KeyError('Weekly frequency not supported')
    else:
        dt = dt.floor(freq)
    
    return dt.strftime(output_format)


"""
Module to plot multiple maps

"""
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from aqua.core.logger import log_configure
from aqua.core.util import plot_box, evaluate_colorbar_limits, cbar_get_label
from .single_map import plot_single_map, plot_single_map_diff
from .styles import ConfigStyle


def plot_maps(maps: list,
              contour: bool = True, sym: bool = False,
              proj: ccrs.Projection = ccrs.Robinson(), extent: list = None,
              style=None, figsize: tuple = None,
              vmin: float = None, vmax: float = None, nlevels: int = 11,
              title: str = None, title_size: int = 16, titles: list = None, titles_size: int = None, 
              cmap='RdBu_r', cbar_label: str = None,
              transform_first=False, cyclic_lon=True,
              return_fig=False, loglevel='WARNING', **kwargs):
    """
    Plot multiple maps.
    This is supposed to be used for maps to be compared together.
    A list of xarray.DataArray objects is expected
    and a map is plotted for each of them

    Args:
        maps (list):          list of xarray.DataArray objects
        contour (bool,opt):   If True, plot a contour map, otherwise a pcolormesh. Defaults to True.
        sym (bool,opt):       symetric colorbar, default is False
        proj (cartopy.crs.Projection,opt): projection, default is ccrs.Robinson()
        extent (list,opt):    extent of the map, default is None
        style (str,opt):      style for the plot, default is the AQUA style
        figsize (tuple,opt):  figure size, default is (6,6) for each map. Here the full figure size is set.
        vmin (float,opt):     minimum value for the colorbar, default is None
        vmax (float,opt):     maximum value for the colorbar, default is None
        nlevels (int,opt):    number of levels for the colorbar, default is 11
        title (str,opt):      super title for the figure
        title_size (int,opt):  size of the super title, default is 16
        titles (list,opt):    list of titles for the maps
        titles_size (int,opt): size of the titles, default is None
        cmap (str,opt):       colormap, default is 'RdBu_r'
        cbar_label (str,opt): colorbar label
        transform_first (bool, optional): If True, transform the data before plotting. Defaults to False.
        cyclic_lon (bool,opt): add cyclic longitude, default is True
        return_fig (bool,opt): return the figure, default is False
        loglevel (str,opt):   log level, default is 'WARNING'
        **kwargs:             Keyword arguments for plot_single_map

    Raises:
        ValueError: if nothing to plot, i.e. maps is None or not a list of xarray.DataArray

    Return:
        fig     if more manipulations on the figure are needed, if return_fig=True
    """
    logger = log_configure(loglevel, 'plot_maps')
    ConfigStyle(style=style, loglevel=loglevel)
    if maps is None or any(not isinstance(data_map, xr.DataArray) for data_map in maps):
        raise ValueError('Maps should be a list of xarray.DataArray')
    else:
        logger.debug('Loading maps')
        maps = [data_map.load(keep_attrs=True) for data_map in maps]

    # Generate the figure
    nrows, ncols = plot_box(len(maps))
    figsize = figsize if figsize is not None else (ncols*6, nrows*5 + 1)
    logger.debug('Creating a %d x %d grid with figsize %s', nrows, ncols, figsize)

    fig = plt.figure(figsize=figsize)

    # Evaluate min and max values for the common colorbar
    if vmin is None or vmax is None or sym:
        vmin, vmax = evaluate_colorbar_limits(maps=maps, sym=sym)

    logger.debug("Setting vmin to %s, vmax to %s", vmin, vmax)

    for i in range(len(maps)):
        logger.debug("Plotting map %d", i)
        fig, ax = plot_single_map(data=maps[i], contour=contour,
                                  proj=proj, extent=extent,
                                  vmin=vmin, vmax=vmax, nlevels=nlevels,
                                  title=titles[i] if titles is not None else None,
                                  title_size=titles_size,
                                  cmap=cmap, cbar=False,
                                  transform_first=transform_first, return_fig=True,
                                  cyclic_lon=cyclic_lon, fig=fig, loglevel=loglevel,
                                  ax_pos=(nrows, ncols, i+1), **kwargs)

    # Adjust the location of the subplots on the page to make room for the colorbar
    fig.subplots_adjust(bottom=0.25, top=0.87, left=0.05, right=0.95,
                        wspace=0.1, hspace=0.2)

    # Add a colorbar axis at the bottom of the graph
    cbar_ax = fig.add_axes([0.2, 0.15, 0.6, 0.03])

    cbar_label = cbar_get_label(data=maps[0], cbar_label=cbar_label, loglevel=loglevel)
    logger.debug('Setting colorbar label to %s', cbar_label)

    mappable = ax.collections[0]

    # Add the colorbar
    cbar = fig.colorbar(mappable, cax=cbar_ax, orientation='horizontal',
                        label=cbar_label)

    # Make the colorbar ticks symmetrical if sym=True
    if sym:
        logger.debug('Setting colorbar ticks to be symmetrical')
        cbar.set_ticks(np.linspace(-vmax, vmax, nlevels + 1))
    else:
        cbar.set_ticks(np.linspace(vmin, vmax, nlevels + 1))

    cbar.ax.ticklabel_format(style='sci', axis='x', scilimits=(-3, 3))

    # Add a super title
    if title:
        logger.debug('Setting super title to %s', title)
        fig.suptitle(title, fontsize=title_size)

    if return_fig:
        return fig


def plot_maps_diff(maps: list,
                   maps_ref: list,
                   contour: bool = True,
                   sym: bool = True, sym_contour: bool = False,
                   proj: ccrs.Projection = ccrs.Robinson(), extent: list = None,
                   style=None, figsize=None,
                   vmin_fill: float = None, vmax_fill: float = None,
                   vmin_contour: float = None, vmax_contour: float = None,
                   nlevels: int = 11, title: str = None, titles: list = None,
                   titles_size: int = None,
                   cmap='RdBu_r', cbar_label: str = None,
                   transform_first=False, cyclic_lon=True,
                   return_fig=False, loglevel='WARNING', **kwargs):
    """
    Plot the difference of multiple maps. This is supposed to be used for maps to be compared together.
    Two lists of xarray.DataArray objects are expected
    and the difference (maps[i] - maps_ref[i]) is plotted for each pair.

    Args:
        maps (list):          list of xarray.DataArray objects
        maps_ref (list):      list of xarray.DataArray reference objects
        contour (bool,opt):   If True, plot a contour map, otherwise a pcolormesh. Defaults to True.
        sym (bool,opt):       symetric colorbar, default is False
        sym_contour (bool,opt): symetric colorbar for the contour, default is False
        proj (cartopy.crs.Projection,opt): projection, default is ccrs.Robinson()
        extent (list,opt):    extent of the map, default is None
        style (str,opt):      style for the plot, default is the AQUA style
        figsize (tuple,opt):  figure size, default is (6,6) for each map. Here the full figure size is set.
        vmin_fill (float,opt):     minimum value for the colorbar of the filled map, default is None
        vmax_fill (float,opt):     maximum value for the colorbar of the filled map, default is None
        vmin_contour (float,opt): minimum value for the colorbar of the contour map, default is None
        vmax_contour (float,opt): maximum value for the colorbar of the contour map, default is None
        nlevels (int,opt):    number of levels for the colorbar, default is 11
        title (str,opt):      super title for the figure
        titles (list,opt):    list of titles for the maps
        titles_size (int,opt): size of the titles, default is None
        cmap (str,opt):       colormap, default is 'RdBu_r'
        cbar_label (str,opt): colorbar label
        transform_first (bool, optional): If True, transform the data before plotting. Defaults to False.
        cyclic_lon (bool,opt): add cyclic longitude, default is True
        return_fig (bool,opt): return the figure, default is False
        loglevel (str,opt):   log level, default is 'WARNING'
        **kwargs:             Keyword arguments for plot_single_map

    Raises:
        ValueError: if nothing to plot, i.e. maps, ref_maps is None or not a list of xarray.DataArray

    Return:
        fig     if more manipulations on the figure are needed, if return_fig=True
    """
    logger = log_configure(loglevel, 'plot_maps_diff')
    ConfigStyle(style=style, loglevel=loglevel)

    if maps is None or any(not isinstance(data_map, xr.DataArray) for data_map in maps)\
            or maps_ref is None or any(not isinstance(data_map, xr.DataArray) for data_map in maps_ref):
        raise ValueError('Maps and reference maps should be lists of xarray.DataArray')
    else:
        logger.debug('Loading maps and reference maps')
        maps = [data_map.load(keep_attrs=True) for data_map in maps]
        maps_ref = [data_map.load(keep_attrs=True) for data_map in maps_ref]

    # Calculate differences
    diffs = [data_map - data_map_ref for data_map, data_map_ref in zip(maps, maps_ref)]

    # Generate the figure
    nrows, ncols = plot_box(len(diffs))
    figsize = figsize if figsize is not None else (ncols*6, nrows*5 + 1)
    logger.debug('Creating a %d x %d grid with figsize %s', nrows, ncols, figsize)

    fig = plt.figure(figsize=figsize)

    # Evaluate min and max values for the common colorbar
    if vmin_fill is None or vmax_fill is None or sym:
        vmin_fill, vmax_fill = evaluate_colorbar_limits(maps=diffs, sym=sym)
    logger.debug("Setting vmin_fill to %s, vmax_fill to %s", vmin_fill, vmax_fill)

    if vmin_contour is None or vmax_contour is None or sym_contour:
        vmin_contour, vmax_contour = evaluate_colorbar_limits(maps=maps, sym=sym_contour)
    logger.debug("Setting vmin_contour to %s, vmax_contour to %s", vmin_contour, vmax_contour)

    for i in range(len(diffs)):
        logger.debug("Plotting map %d", i)
        fig, ax = plot_single_map_diff(data=maps[i], data_ref=maps_ref[i], contour=contour,
                                       proj=proj, extent=extent, style=style,
                                       vmin_fill=vmin_fill, vmax_fill=vmax_fill, nlevels=nlevels,
                                       title=titles[i] if titles is not None else None,
                                       title_size=titles_size,
                                       cmap=cmap, cbar=False,
                                       sym=False, sym_contour=False,
                                       transform_first=transform_first, return_fig=True,
                                       cyclic_lon=cyclic_lon, fig=fig, loglevel=loglevel,
                                       ax_pos=(nrows, ncols, i+1), **kwargs)

    # Adjust the location of the subplots on the page to make room for the colorbar
    fig.subplots_adjust(bottom=0.25, top=0.9, left=0.05, right=0.95,
                        wspace=0.1, hspace=0.2)

    # Add a colorbar axis at the bottom of the graph
    cbar_ax = fig.add_axes([0.2, 0.15, 0.6, 0.03])

    cbar_label = cbar_get_label(data=diffs[0], cbar_label=cbar_label, loglevel=loglevel)
    logger.debug('Setting colorbar label to %s', cbar_label)

    mappable = ax.collections[0]

    # Add the colorbar
    cbar = fig.colorbar(mappable, cax=cbar_ax, orientation='horizontal',
                        label=cbar_label)

    # Make the colorbar ticks symmetrical if sym=True
    if sym:
        logger.debug('Setting colorbar ticks to be symmetrical')
        cbar.set_ticks(np.linspace(-vmax_fill, vmax_fill, nlevels + 1))
    else:
        cbar.set_ticks(np.linspace(vmin_fill, vmax_fill, nlevels + 1))

    cbar.ax.ticklabel_format(style='sci', axis='x', scilimits=(-3, 3))

    # Add a super title
    if title:
        logger.debug('Setting super title to %s', title)
        fig.suptitle(title, fontsize=16)

    if return_fig:
        return fig



"""Test for some of the utils"""

import pytest
import xarray as xr
import numpy as np
import pandas as pd
from aqua import Reader
from aqua.core.util import extract_literal_and_numeric, file_is_complete, to_list, convert_data_units
from aqua.core.util import format_realization, extract_attrs, time_to_string
from aqua.core.util.string import strlist_to_phrase, lat_to_phrase
from aqua.core.util.units import multiply_units
from conftest import LOGLEVEL

@pytest.fixture
def test_text():
    return [
        ("1D", ("D", 1)),
        ("MS", ("MS", 1)),
        ("300h", ("h", 300)),
        ("", (None, None)),
    ]

@pytest.mark.aqua
def test_extract_literal_and_numeric(test_text):
    for input_text, expected_output in test_text:
        result = extract_literal_and_numeric(input_text)
        assert result == expected_output

loglevel = LOGLEVEL

@pytest.mark.aqua
def test_convert_data_units():
    """Test the check_data function"""
    data = Reader(catalog='ci', model='ERA5', exp='era5-hpz3', source='monthly', loglevel=loglevel).retrieve()
    initial_units = data['tprate'].attrs['units']

    # Dataset test
    data_test = convert_data_units(data=data, var='tprate', units='mm/day', loglevel=loglevel)
    assert data_test['tprate'].attrs['units'] == 'mm/day'

    # DataArray test
    data = data['tprate']
    data_test = convert_data_units(data=data, var='tprate', units='mm/day', loglevel=loglevel)
    # We don't test values since this is done in the test_fixer.py
    assert data_test.attrs['units'] == 'mm/day'
    assert f"Converting units of tprate: from {initial_units} to mm/day" in data_test.attrs['history']

    # Test with no conversion to be done
    data_test = convert_data_units(data=data, var='tprate', units=initial_units, loglevel=loglevel)
    assert data_test.attrs['units'] == initial_units
    assert f"Converting units of tprate: from {initial_units} to mm/day" not in data_test.attrs['history']


# Define a fixture to create a sample netCDF file for testing
@pytest.mark.aqua
class TestFileIsComplete:
    """The File is Complete testing class"""

    @pytest.fixture
    def sample_netcdf(self, tmp_path):
        """Create a sample Dataset and its file"""
        filename = tmp_path / "sample_netcdf.nc"
        data = xr.DataArray(np.random.rand(3, 4, 5), dims=("time", "lat", "lon"))
        time_values = [pd.Timestamp("2024-01-01") + pd.Timedelta(days=i) for i in range(3)]
        data = data.assign_coords(time=time_values)
        data.name = "sample_data"
        dataset = xr.Dataset({"sample_data": data})
        dataset.to_netcdf(filename)
        return filename
    
    def test_file_is_complete_existing_file(self, sample_netcdf):
        result = file_is_complete(sample_netcdf)
        assert result is True

    @pytest.mark.parametrize("mindate,expected_result", [("2023-12-31", False), ("2025-01-01", True)])
    def test_file_is_complete_full_nan_with_mindate(self, tmp_path, mindate, expected_result):
        filename = tmp_path / "sample_netcdf.nc"
        data = xr.DataArray(np.random.rand(3, 4, 5), dims=("time", "lat", "lon"))
        time_values = [pd.Timestamp("2024-01-01") + pd.Timedelta(days=i) for i in range(3)]
        data = data.assign_coords(time=time_values)
        data.name = "sample_data"
        data[:,:,:] = np.nan
        dataset = xr.Dataset({"sample_data": data})
        dataset["sample_data"].attrs["mindate"] = mindate
        dataset.to_netcdf(filename)
        result = file_is_complete(filename, loglevel='info')
        assert result == expected_result
    
    @pytest.mark.parametrize("mindate,expected_result", [("2023-12-31", False), ("2024-02-01", True)])
    def test_file_is_complete_partial_nan_with_mindate(self, tmp_path, mindate, expected_result):
        filename = tmp_path / "sample_netcdf.nc"
        data = xr.DataArray(np.random.rand(3, 4, 5), dims=("time", "lat", "lon"))
        time_values = [pd.Timestamp("2024-01-01") + pd.Timedelta(days=i * 40) for i in range(3)]
        data = data.assign_coords(time=time_values)
        data.name = "sample_data"
        data[0,:,:] = np.nan
        dataset = xr.Dataset({"sample_data": data})
        dataset["sample_data"].attrs["mindate"] = mindate
        dataset.to_netcdf(filename)
        result = file_is_complete(filename, loglevel='info')
        assert result == expected_result

    def test_file_is_complete_nonexistent_file(self, tmp_path):
        non_existent_file = tmp_path / "non_existent.nc"
        result = file_is_complete(non_existent_file)
        assert result is False

    def test_file_is_complete_empty_file(self, tmp_path):
        empty_file = tmp_path / "empty.nc"
        xr.DataArray().to_netcdf(empty_file)
        result = file_is_complete(empty_file)
        assert result is False

    def test_file_is_complete_nan_file(self, tmp_path):
        nan_file = tmp_path / "nan.nc"
        nan_data = xr.DataArray(np.full((3, 4, 5), np.nan), dims=("time", "lat", "lon"))
        nan_data.to_netcdf(nan_file)
        result = file_is_complete(nan_file)
        assert result is False
        #assert "full of NaN" in caplog.text

    def test_file_is_complete_with_missing_time(self, tmp_path):
        valid_with_missing_time_file = tmp_path / "valid_with_missing_time.nc"
        data = xr.DataArray(np.random.rand(3, 4, 5), dims=("time", "lat", "lon"))
        data[0,:,:] = np.nan # Introduce NaN value
        data.to_netcdf(valid_with_missing_time_file)
        result = file_is_complete(valid_with_missing_time_file)
        assert result is False

    def test_file_is_complete_valid_with_nan(self, tmp_path):
        valid_with_nan_file = tmp_path / "valid_with_nan.nc"
        data = xr.DataArray(np.random.rand(3, 4, 5), dims=("time", "lat", "lon"))
        data[:,1,:] = np.nan # Introduce NaN value
        data.to_netcdf(valid_with_nan_file)
        result = file_is_complete(valid_with_nan_file)
        assert result is True

@pytest.mark.parametrize("arg, expected", [
    (None, []),                        # Test None
    ([1, 2, 3], [1, 2, 3]),              # Test list (unchanged)
    ((4, 5), [4, 5]),                    # Test tuple
    ({6, 7}, [6, 7]),                    # Test set
    ({"capa": 1, "tonda": 2}, ["capa", "tonda"]),      # Test dictionary (keys)
    ("maccio", ["maccio"]),# Test string (split into chars)
    (8, [8]),                            # Test single integer
    (3.14, [3.14]),                      # Test single float
    (True, [True]),                      # Test single boolean
    ([], []),                            # Test empty list
    ({}, []),                            # Test empty dictionary
    (set(), []),                         # Test empty set
    ((), []),                            # Test empty tuple
])
def test_to_list(arg, expected):
    assert to_list(arg) == expected

@pytest.mark.aqua
def test_format_realization():
    """Test the format_realization function"""
    assert format_realization() == "r1"
    assert format_realization(1) == "r1"
    assert format_realization("2") == "r2"
    assert format_realization("test") == "test"
    assert format_realization("") == "r1"
    assert format_realization(None) == "r1"
    assert format_realization(["1", "2", "test", 3, None]) == ["r1", "r2", "test", "r3", "r1"]

@pytest.mark.aqua
def test_extract_attrs():
    assert extract_attrs(None, "attr") is None # Data is None
    ds_with_attr = xr.Dataset()
    ds_with_attr.attrs = {"attr": "value1"}
    ds_without_attr = xr.Dataset()
    assert extract_attrs(ds_with_attr, "attr") == "value1" # Single dataset with attribute
    assert extract_attrs(ds_without_attr, "attr") is None # Single dataset without attribute
    result = extract_attrs([ds_with_attr, ds_without_attr], "attr") 
    assert result == ["value1", None] # List of datasets

@pytest.mark.aqua
def test_strlist_to_phrase():
    """Test the strlist_to_phrase function"""
    # Test empty list
    assert strlist_to_phrase([]) == ""
    # Test single item
    assert strlist_to_phrase(["A"]) == "A"
    assert strlist_to_phrase(["Hello"]) == "Hello"
    # Test two items
    assert strlist_to_phrase(["A", "B"]) == "A and B"
    assert strlist_to_phrase(["Hello", "World"]) == "Hello and World"
    # Test more than three items
    assert strlist_to_phrase(["A", "B", "C", "D"]) == "A, B, C and D"
    assert strlist_to_phrase(["A", "B", "C", "D"], oxford_comma=True) == "A, B, C, and D"
    # Test with mixed content
    assert strlist_to_phrase(["Mod", "Exp", "Source"]) == "Mod, Exp and Source"
    assert strlist_to_phrase(["Mod", "Exp", "Source"], oxford_comma=True) == "Mod, Exp, and Source"

@pytest.mark.aqua
def test_lat_to_phrase():
    """Test the lat_to_phrase function"""
    # Test northern latitudes
    assert lat_to_phrase(90) == "90°N"
    # Test southern latitudes
    assert lat_to_phrase(-1) == "1°S"
    # Test 0 latitude
    assert lat_to_phrase(0) == "0°N"
    
# Uncomment this test if the flip_time function is uncommented in aqua/util/coord.py
# def test_flip_time():
#     """Test the flip_time function"""
#     time = np.arange(0, 10)
#     data = xr.DataArray(np.random.rand(10, 3, 4), dims=("time", "lat", "lon"))
#     data = data.assign_coords(time=time)
#     flipped_data = flip_time(data)
#     assert flipped_data.time.values[0] == 9
#     assert flipped_data.time.values[-1] == 0

@pytest.mark.aqua
@pytest.mark.parametrize("unit1, unit2, expected", [
    ("m", "m2", "meter ** 3"),                         # Basic area × length = volume
    ("m", "m", "meter ** 2"),                          # Length × length = area
    ("kg", "m/s2", "kilogram * meter / second ** 2"),  # Mass × acceleration = force
    ("m/s", "s", "meter"),                             # Velocity × time = distance
    ("1", "m", "meter"),                               # Dimensionless × length
    ("kg/m3", "m3", "kilogram"),                       # Density × volume = mass
])
def test_multiply_units(unit1, unit2, expected):
    """Test the multiply_units function with various unit combinations"""
    result = multiply_units(unit1, unit2)
    assert result == expected

@pytest.mark.aqua
def test_multiply_units_no_base_conversion():
    """Test multiply_units without converting to base units"""
    result = multiply_units("km", "km", to_base_units=False)
    assert result == "kilometer ** 2"

@pytest.mark.aqua
def test_multiply_units_no_normalization():
    """Test multiply_units without normalization"""
    result = multiply_units("m", "m", normalise_units=False)
    assert result == "meter ** 2"



"""Testing for the catalog generator"""

import subprocess
import os
import pytest
import logging
from pathlib import Path
from aqua.core.util import load_yaml, dump_yaml
from aqua.core.console.catgen import AquaFDBGenerator
from conftest import LOGLEVEL

loglevel = LOGLEVEL

def load_and_prepare(tmp_path, model, kind, reso, num_of_realizations=1):
    """
    Load configuration, execute catgen, and return generated catalog sources.
    
    This function ensures test isolation by using tmp_path for the catalog directory,
    preventing race conditions when tests run in parallel. All generated files are
    automatically cleaned up by pytest's tmp_path fixture.

    Args:
        tmp_path: Temporary directory provided by pytest (Path or str)
        model: Model to be checked (IFS-NEMO, IFS-FESOM, ICON)
        kind: Data portfolio type (minimal, reduced, full)
        reso: Resolution of the data (lowres, intermediate, production, etc.)
        num_of_realizations: Number of realizations for the ensemble
        
    Returns:
        dict: Catalog sources dictionary loaded from the generated YAML file
        
    Raises:
        subprocess.CalledProcessError: If catgen command fails
        AssertionError: If generated catalog files are not found
    """
    tmp_path = Path(tmp_path)
    config_template = Path('tests/catgen/config-test-catgen.j2')
    
    # Prepare configuration
    config = _prepare_config(config_template, model, kind, reso, num_of_realizations, tmp_path)
    
    # Setup isolated catalog directory structure
    _setup_catalog_directory(config, tmp_path)
    
    # Execute catgen command
    _run_catgen(config, kind, tmp_path)
    
    # Load and return generated sources
    return _load_generated_sources(config)


def _prepare_config(template_path, model, kind, reso, num_of_realizations, tmp_path):
    """Prepare configuration from template with test-specific values."""
    definitions = {
        'model': model,
        'kind': kind,
        'resolution': reso,
        'num_of_realizations': num_of_realizations,
        'expid': 'test'
    }
    config = load_yaml(str(template_path), definitions)
    
    # Use tmp_path for catalog directory to ensure test isolation
    catalog_base = tmp_path / 'Climate-DT-catalog'
    config['repos']['Climate-DT-catalog_path'] = str(catalog_base)
    
    return config


def _setup_catalog_directory(config, tmp_path):
    """Create isolated catalog directory structure and required files."""
    catalog_base = Path(config['repos']['Climate-DT-catalog_path'])
    catalog_dir = config['catalog_dir']
    catalog_yaml_path = catalog_base / 'catalogs' / catalog_dir / 'catalog.yaml'
    
    # Create directory structure
    catalog_yaml_path.parent.mkdir(parents=True, exist_ok=True)
    
    # Create initial catalog.yaml if it doesn't exist (required by catgen)
    if not catalog_yaml_path.exists():
        dump_yaml(str(catalog_yaml_path), {'sources': {}})


def _run_catgen(config, kind, tmp_path):
    """Execute the catgen command via subprocess."""
    config_path = tmp_path / 'test.yaml'
    dump_yaml(str(config_path), config)
    
    command = ["aqua", "catgen", '-p', kind, '-c', str(config_path), '-l', loglevel]
    
    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        logging.info("Command succeeded with output: %s", result.stdout)
    except subprocess.CalledProcessError as e:
        logging.error("Command failed with error: %s", e)
        logging.error("Return code: %s", e.returncode)
        logging.error("stderr: %s", e.stderr)
        raise
    except Exception as e:
        logging.error("Unexpected error: %s", e)
        raise


def _load_generated_sources(config):
    """Load and validate the generated catalog sources."""
    catalog_path = Path(config['repos']['Climate-DT-catalog_path'])
    catalog_dir = config['catalog_dir']
    model = config['model']
    exp = config['exp']
    
    catalog_entry_dir = catalog_path / 'catalogs' / catalog_dir / 'catalog' / model.upper()
    entry_file = catalog_entry_dir / f'{exp}.yaml'
    main_yaml_file = catalog_entry_dir / 'main.yaml'
    
    # Validate that required files were generated
    assert main_yaml_file.exists(), f"main.yaml not found at {main_yaml_file}"
    assert entry_file.exists(), f"Catalog entry not found at {entry_file}"
    
    return load_yaml(str(entry_file))

@pytest.mark.parametrize(('model,nsources,nocelevels'),
                        [('IFS-NEMO', 4, 75)])
@pytest.mark.catgen
def test_catgen_minimal(tmp_path, model, nsources, nocelevels):
    """test for minimal portfolio"""

    ensemble = 5 

    sources = load_and_prepare(tmp_path=tmp_path, model=model,
                               kind='minimal', reso='lowres',
                               num_of_realizations=ensemble)

    # check how many sources
    assert len(sources['sources']) == nsources

    # check if realization is correctly formatted
    assert "realization: '{{ realization }}'"

    # check number of vertical levels in the atmosphere
    assert len(sources['sources'][f'monthly-hpz5-atm3d']['metadata']['levels']) == 19

    # check number of vertical levels in the ocean
    assert len(sources['sources'][f'monthly-hpz5-oce3d']['metadata']['levels']) == nocelevels

    # check ensembles are correctly produced
    assert sources['sources'][f'monthly-hpz5-atm3d']['parameters']['realization']['allowed'] == [*range(1, ensemble+1)]


@pytest.mark.parametrize(('model,nsources,nocelevels'),
                        [('IFS-NEMO', 9, 75)])
                        # ('IFS-FESOM', 5, 47)])
@pytest.mark.catgen
def test_catgen_reduced(tmp_path, model, nsources, nocelevels):
    """test for reduced portfolio"""

    ensemble = 5

    sources = load_and_prepare(tmp_path=tmp_path, model=model,
                               kind='reduced', reso='intermediate',
                               num_of_realizations=ensemble)

    # check how many sources
    assert len(sources['sources']) == nsources

    # check if realization is correctly formatted
    assert "realization: '{{ realization }}'"

    # check number of vertical levels in the atmosphere
    if model == 'IFS-NEMO':
        grid, freq = 'hpz7', 'monthly'
    #elif model == 'IFS-FESOM':
    #   grid, freq = 'hpz7', 'daily'
    else:
        raise ValueError(f'{model} not supported!')
    assert len(sources['sources'][f'monthly-{grid}-atm3d']['metadata']['levels']) == 19

    # check number of vertical levels in the atmosphere
    assert len(sources['sources'][f'{freq}-{grid}-oce3d']['metadata']['levels']) == nocelevels

    # check ensembles are correctly produced
    assert sources['sources'][f'monthly-{grid}-atm3d']['parameters']['realization']['allowed'] == [*range(1, ensemble+1)]


@pytest.mark.parametrize(('model,nsources,nocelevels'),
                        [('IFS-NEMO', 28, 75),
                         ('IFS-FESOM', 31, 69),
                         ('ICON', 27, 72)])
@pytest.mark.catgen
def test_catgen_full(tmp_path, model, nsources, nocelevels):
    """test for full portfolio"""

    sources = load_and_prepare(tmp_path, model, 'full', 'production')

    # check how many sources
    assert len(sources['sources']) == nsources

    # check number of vertical levels in the atmosphere
    assert len(sources['sources']['hourly-hpz10-atm3d']['metadata']['levels']) == 19

    # check number of vertical levels in the atmosphere
    assert len(sources['sources']['daily-hpz10-oce3d']['metadata']['levels']) == nocelevels


MANDATORY_KEYS_TO_TEST = [
    "author",
    "model",
    "repos.data-portfolio_path",
]
@pytest.mark.parametrize("missing_key", MANDATORY_KEYS_TO_TEST)

@pytest.mark.catgen
def test_catgen_missing_key(tmp_path, missing_key):
    """
    Ensure that a ValueError is raised when a required configuration key is missing.
    """
    config_path = 'tests/catgen/config-test-catgen.j2'
    config = load_yaml(config_path)

    if missing_key.startswith("repos."):
        subkey = missing_key.split(".", 1)[1]
        if "repos" in config and subkey in config["repos"]:
            del config["repos"][subkey]
    else:
        config.pop(missing_key, None)

    dump_path = os.path.join(tmp_path, "test.yaml")
    dump_yaml(dump_path, config)

    with pytest.raises(ValueError, match="Missing required configuration keys"):
        AquaFDBGenerator(config_path=dump_path, data_portfolio="minimal")



"""Test cases for the Trender class."""

import pytest
from aqua import Reader
from conftest import LOGLEVEL

loglevel = LOGLEVEL

@pytest.mark.aqua
class TestTrender:
    """Test class for Trender functionality."""

    @pytest.fixture(scope='class')
    def reader(self):
        return Reader(model="IFS", exp="test-tco79", source='long', loglevel=loglevel)

    @pytest.fixture(scope='class')
    def data(self, reader):
        """Retrieve all data once for all tests in this class"""
        return reader.retrieve()

    def test_coeffs_dataset(self, reader, data):
        """Test for polynomial coefficients on Dataset"""
        block1 = data.isel(time=slice(0, 1000))
        coeffs = reader.trender.coeffs(block1, degree=1)
        avg = coeffs['2t'].sel(degree=1).mean().values
        assert float(avg) == pytest.approx(-7.91903731e-17, rel=1e-5)
        coeffs = reader.trender.coeffs(block1, degree=1, normalize=True)
        avg = coeffs['2t'].sel(degree=1).mean().values
        assert float(avg) == pytest.approx(-0.0002850853431, rel=1e-5)

    def test_trend_dataarray(self, reader, data):
        """Trivial test for trend on DataArray"""
        block1 = data['2t'].isel(time=slice(0, 1000))
        trend1 = reader.trender.trend(block1).aqua.fldmean()

        assert trend1.shape == (1000,)
        assert pytest.approx(trend1.values[300]) == 285.908

    def test_detrend_dataarray(self, reader, data):
        """Trivial test for detrending on DataArray"""
        block1 = data['2t'].isel(time=slice(0, 1000))
        det1 = reader.detrend(block1).aqua.fldmean()

        assert det1.shape == (1000,)
        assert pytest.approx(det1.values[300]) == 0.3778275

    def test_detrend_dataset(self, reader, data):
        """Second trivial test for detrending on Dataset"""
        block2 = data[['2t', 'skt']].isel(time=slice(0, 100))
        det2 = reader.detrend(block2, dim='time', degree=2)

        assert list(det2.data_vars) == ['2t', 'skt']
        assert pytest.approx(det2['skt'].isel(time=10, lon=2, lat=2).values) == -0.098381225331


"""Class to load the choosen style for graphical utilities."""
import os
import matplotlib.pyplot as plt
from aqua.core.logger import log_configure
from aqua.core.configurer import ConfigPath
from aqua.core.util import load_yaml


class ConfigStyle(ConfigPath):
    """Class to load the choosen style for graphical utilities."""
    def __init__(self,
                 style: str = None,
                 filename: str = 'config-aqua.yaml',
                 configdir: str = None,
                 loglevel: str = 'WARNING'):
        """Initialize the class.

        Args:
            style (str): name of the style to load.
                     If not provided, it will be read from the configuration file.
            filename (str): name of the configuration file. Default is 'config-aqua.yaml'.
            configdir (str): path to the configuration directory.
                            If not provided, it is determined by the `get_config_dir` method.
            loglevel (str): logging level. Default is 'WARNING'
        """

        # Initialize the ConfigPath class
        super().__init__(configdir=configdir, filename=filename, loglevel=loglevel)
        self.logger = log_configure(log_level=loglevel, log_name="ConfigStyle")

        if style is not None:
            self.style = style
        else:  # Read the style from the configuration file
            configfile = load_yaml(self.config_file)
            if 'options' in configfile:
                self.style = configfile['options'].get('style', 'aqua')
            else:
                self.style = 'aqua'

        style_dir = os.path.join(self.configdir, 'styles')
        filename = self.style + '.mplstyle'
        self.style_file = os.path.join(style_dir, filename)
        self.logger.debug("Style file: %s", self.style_file)
        self.load_style()

    def load_style(self):
        """Load the choosen style."""
        try:
            plt.style.use(self.style_file)
            self.logger.debug("Setting style %s from file %s", self.style, self.style_file)
        except OSError:
            plt.style.use(self.style)
            self.logger.debug("Setting matplotlib style %s", self.style)